#### SEARCH BY TRADING - B3

import pandas as pd
import numpy as np
from zipfile import ZipFile
from stpstone.settings._global_slots import YAML_B3
from stpstone.loggs.create_logs import CreateLog
from stpstone.handling_data.lists import HandlingLists
from stpstone.cals.handling_dates import DatesBR
from stpstone.handling_data.json import JsonFiles
from stpstone.handling_data.folders import DirFilesManagement
from stpstone.handling_data.txt import HandlingTXTFiles
from stpstone.handling_data.str import StrHandler
from stpstone.finance.b3.up2data_web import UP2DATAB3
from stpstone.handling_data.html import HtmlHndler
from stpstone.handling_data.xml import XMLFiles
from stpstone.finance.derivatives.options.european import EuropeanOptions
from stpstone.finance.b3.market_data import MDB3
from stpstone.loggs.db_logs import DBLogs


class TradingFilesB3:

    def __init__(self, int_wd_bef:int=1, str_fillna:str='-1', str_fillna_dt:str='2100-12-31', 
                 str_fillna_ts:str='2100-12-31 00:00:00', str_fmt_dt:str='YYYY-MM-DD') -> None:
        self.int_wd_bef = int_wd_bef
        self.str_fillna = str_fillna
        self.str_fillna_dt = str_fillna_dt
        self.str_fillna_ts = str_fillna_ts
        self.str_fmt_dt = str_fmt_dt

    def price_report(self, bl_debug=False):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # setting variables
        list_ser = list()
        # price report file .zip
        url = YAML_B3['price_report'][
            'url'].format(DatesBR().sub_working_days(DatesBR().curr_date,
                                                     self.int_wd_bef).strftime('%y%m%d'))
        # mtm b3 on memry
        zipfile = DirFilesManagement().get_zip_from_web_in_memory(
            url, bl_io_interpreting=YAML_B3[
                'price_report']['bl_io_interpreting'],
            bl_verify=YAML_B3['price_report']['bl_verify']
        )
        # dealing with nested zip file
        zipfile = ZipFile(zipfile)
        # iterate through the files in the zip
        for file_info in zipfile.infolist():
            #   opennig just xml files
            if file_info.filename.endswith('.xml'):
                xml_file = zipfile.open(file_info)
                xml_file = xml_file.read()
            else:
                continue
            #   soup parser to xml trading options
            soup_xml = XMLFiles().xml_memory_parser(xml_file)
            #   looping through tags of BOV & BMF markets
            for key, value in YAML_B3['price_report']['tags'].items():
                #   finding key of interest within soup
                for soup_content in soup_xml.find_all(key):
                    #   reseting variables
                    dict_ = dict()
                    #   adding all tags availabe to the dictionary
                    for tag in value:
                        try:
                            #   get tag in order to find attributes of interest
                            tag_ = soup_content.find(tag)
                            if bl_debug == True:
                                print(tag_)
                            dict_[tag] = tag_.get_text()
                            #   getting currency of notionals
                            if YAML_B3['price_report']['attrb_currency'] in tag_.attrs:
                                dict_[YAML_B3['price_report']['col_currency']] = tag_.attrs[
                                    YAML_B3['price_report']['attrb_currency']]
                        except AttributeError:
                            continue
                    #   appendig dictionary to list
                    list_ser.append(dict_)
        # uploading list of dictionaries to pandas
        df_pr = pd.DataFrame(list_ser)
        # removing duplicates
        df_pr.drop_duplicates(inplace=True)
        # adding logging
        df_pr = DBLogs().audit_log(
            df_pr, 
            url,
            DatesBR().utc_from_dt(
                DatesBR().sub_working_days(DatesBR().curr_date, self.int_wd_bef)
            )
        )
        # returning dataframe
        return df_pr

    @property
    def tradable_securities(self):
        '''
        DOCSTRING: TRADABLE SECURITIES IN BRAZILLIAN EXCHANGE MARKET
        INPUTS: WORKING DAYS BEFORE
        OUTPUTS: DATAFRAME
        '''
        # setting variables
        list_ser = list()
        # url
        url = YAML_B3['securities_tradable']['url'].format(DatesBR().sub_working_days(
            DatesBR().curr_date, self.int_wd_bef).strftime('%y%m%d'))
        # zip to memory
        zipfile = DirFilesManagement().get_zip_from_web_in_memory(
            url, bl_io_interpreting=YAML_B3[
                'securities_tradable']['bl_io_interpreting'],
            bl_verify=YAML_B3['securities_tradable']['bl_verify']
        )
        # extracting content
        zipfile = ZipFile(zipfile)
        # iterate through the files in the zip
        for file_info in zipfile.infolist():
            #   opennig just xml files
            if file_info.filename.endswith('.json'):
                json_file = zipfile.open(file_info)
                json_file = json_file.read()
            else:
                continue
            #   reading json file
            json_trd_sec = JsonFiles().loads_message_like(json_file)
            #   import to serialized list
            list_ser.extend(json_trd_sec)
        # appending list serialized to dataframe
        df_ = pd.DataFrame(list_ser)
        # filling na, in order to change data types
        df_.fillna(value={
            "MaturityDate": self.str_fillna_dt,
            "IssueDate": self.str_fillna_dt,
            "SettlDate": self.str_fillna_dt,
            "SecurityValidityTimestamp": self.str_fillna_ts
        }, inplace=True)
        df_.fillna(self.str_fillna, inplace=True)
        # changing data types
        df_ = df_.astype({
            "Symbol": str,
            "SecurityID": np.int64,
            "SecurityIDSource": int,
            "SecurityExchange": str,
            "NoApplIDs": str,
            "Product": int,
            "CFICode": str,
            "SecurityGroup": str,
            "SecurityType": str,
            "SecuritySubType": str,
            "MaturityMonthYear": str,
            "MaturityDate": str,
            "IssueDate": str,
            "CountryOfIssue": str,
            "ContractMultiplier": float,
            "SecurityDesc": str,
            "ContractSettlMonth": str,
            "SettlType": str,
            "SettlDate": str,
            "MinPriceIncrement": float,
            "TickSizeDenominator": int,
            "MinOrderQty": float,
            "MaxOrderQty": np.int64,
            "NoInstrAttrib": str,
            "NoUnderlyings": str,
            "Currency": str,
            "SettlCurrency": str,
            "Asset": str,
            "SecurityValidityTimestamp": str,
            "MarketSegmentID": int,
            "NoLotTypeRules": str,
            "ImpliedMarketIndicator": int,
            "MinCrossQty": int,
            "ClearingHouseID": str,
            "ISINNumber": str,
            "MultiLegModel": str,
            "MultiLegPriceMethod": str,
            "SecurityStrategyType": str,
            "NoLegs": str,
            "NoSharesIssued": str,
            "GovernanceIndicator": str,
            "CorporateActionEventID": int,
            "PriceDivisor": str,
            "PutOrCall": str,
            "StrikePrice": str,
            "StrikeCurrency": str,
            "ExerciseStyle": str,
            "SecurityMatchType": str
        })
        for col_ in [
            "MaturityDate",
            "IssueDate",
            "SettlDate"
        ]:
            df_[col_] = [DatesBR().str_date_to_datetime(d, self.str_fmt_dt) for d in df_[col_]]
        for col_ in [
            "SecurityValidityTimestamp"
        ]:
            df_[col_] = [DatesBR().timestamp_separator_string_to_datetime(
                d, substring_datetime=' ', format=self.str_fmt_dt) for d in df_[col_]]
        # adding logging
        df_ = DBLogs().audit_log(
            df_, 
            url,
            DatesBR().utc_from_dt(
                DatesBR().sub_working_days(DatesBR().curr_date, self.int_wd_bef)
            )
        )
        # retornando dataframe de interesse
        return df_

    @property
    def daily_liquidity_limits(self):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # url
        url = YAML_B3['daily_liquidity_limits']['url'].format(DatesBR().sub_working_days(
            DatesBR().curr_date, self.int_wd_bef).strftime('%y%m%d'))
        # zip to memory
        zipfile = DirFilesManagement().get_zip_from_web_in_memory(
            url, bl_io_interpreting=YAML_B3[
                'daily_liquidity_limits']['bl_io_interpreting'],
            bl_verify=YAML_B3['daily_liquidity_limits']['bl_verify'])
        # extracting content
        zipfile = ZipFile(zipfile)
        # popping file name
        list_zip_files = zipfile.namelist()
        file_name = list_zip_files.pop()
        # extracting file of interest
        name_file_extract = zipfile.open(file_name)
        # open txt in dataframe
        reader = pd.read_csv(
            name_file_extract, sep=';', skiprows=1, names=[
                YAML_B3['daily_liquidity_limits']['col_market'],
                YAML_B3['daily_liquidity_limits']['col_instrument_origin'],
                YAML_B3['daily_liquidity_limits']['col_isin'],
                YAML_B3['daily_liquidity_limits']['col_symbol'],
                YAML_B3['daily_liquidity_limits']['col_dll'],
            ], usecols=[0, 1, 2, 3, 4]
        )
        df_dll = pd.DataFrame(reader)
        # altering column types
        df_dll = df_dll.astype({
            YAML_B3['daily_liquidity_limits']['col_market']: str,
            YAML_B3['daily_liquidity_limits']['col_instrument_origin']: str,
            YAML_B3['daily_liquidity_limits']['col_isin']: np.int64,
            YAML_B3['daily_liquidity_limits']['col_symbol']: str,
            YAML_B3['daily_liquidity_limits']['col_dll']: float
        })
        # adding logging
        df_dll = DBLogs().audit_log(
            df_dll, 
            url,
            DatesBR().utc_from_dt(
                DatesBR().sub_working_days(DatesBR().curr_date, self.int_wd_bef)
            )
        )
        # retornando dataframe de interesse
        return df_dll

    def options_b3(self, float_value_error=-1000.0, bl_debug=False, 
                   float_days_year=365.0):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # setting variables
        list_ser = list()
        # raw maximum theoretical margins
        list_rows_mtm = self.carga_mtm_b3
        # b3 financial indicators
        df_fin_ind = MDB3().financial_indicators_b3
        # risk-free rate
        r = df_fin_ind[df_fin_ind[YAML_B3['financial_indicators']['col_si_code']] \
                       == YAML_B3['financial_indicators']['int_si_code_selic']][
                           YAML_B3['financial_indicators']['col_rate']
                       ].tolist()[0]
        if bl_debug == True:
            print('RISK-FREE RATE: {}'.format(r))
        # options traded in b3 exchange - in zip file
        url = YAML_B3['options_traded_b3'][
            'url'].format(DatesBR().sub_working_days(DatesBR().curr_date,
                                                     self.int_wd_bef).strftime('%y%m%d'))
        # carga para a memória margens teóricas máximas b3
        zipfile = DirFilesManagement().get_zip_from_web_in_memory(
            url, bl_io_interpreting=YAML_B3[
                'options_traded_b3']['bl_io_interpreting'],
            bl_verify=YAML_B3['options_traded_b3']['bl_verify']
        )
        # retorno é um novo zip com dois xmls
        zipfile = ZipFile(zipfile)
        # iterate through the files in the zip
        for file_info in zipfile.infolist():
            #   opennig just xml files
            if file_info.filename.endswith('.xml'):
                xml_file = zipfile.open(file_info)
                xml_file = xml_file.read()
            else:
                continue
            #   soup parser to xml trading options
            soup_xml = XMLFiles().xml_memory_parser(xml_file)
            #   looping through tags of BOV & BMF markets
            for key, value in YAML_B3['options_traded_b3']['tags'].items():
                #   finding key of interest within soup
                for soup_content in soup_xml.find_all(key):
                    #   reseting variables
                    dict_ = dict()
                    bl_break_opt_price = False
                    bl_break_spt_price = False
                    #   adding all tags availabe to the dictionary
                    for tag in value:
                        dict_[tag] = soup_content.find(tag).get_text()
                        if bl_debug == True:
                            print(soup_content.find(tag).get_text())
                    if bl_debug == True:
                        print('ISIN: {}'.format(soup_content.find(YAML_B3['options_traded_b3'][
                            'keys']['isin']).get_text()))
                        print('TICKER: {}'.format(dict_[YAML_B3['options_traded_b3']['keys'][
                            'ticker']]))
                    #   spot ticker
                    str_isin = soup_content.find(YAML_B3['options_traded_b3']['keys']['isin']).get_text()
                    dict_[YAML_B3['options_traded_b3']['keys']['spot']] = \
                        str_isin[2:6] + str_isin[6].replace('9', '11')
                    if bl_debug == True:
                        print('SPOT TICKER: {}'.format(dict_[YAML_B3['options_traded_b3'][
                            'keys']['spot']]))
                    #   calendar days to expiration
                    dict_[YAML_B3['options_traded_b3']['keys']['cd_exp']] = \
                        DatesBR().delta_calendar_days(
                            DatesBR().sub_working_days(
                                DatesBR().curr_date, abs(-self.int_wd_bef-1)), 
                                DatesBR().str_date_to_datetime(str(dict_[
                                        YAML_B3['options_traded_b3']['keys']['xpr_dt']
                                    ]), YAML_B3['options_traded_b3']['dt_input_format']
                            )
                        )
                    if bl_debug == True:
                        print('CALENDAR DAYS TO EXPIRATION: {}'.format(
                                DatesBR().delta_calendar_days(
                                DatesBR().sub_working_days(
                                    DatesBR().curr_date, abs(-self.int_wd_bef-1)), 
                                    DatesBR().str_date_to_datetime(str(dict_[
                                            YAML_B3['options_traded_b3']['keys']['xpr_dt']
                                        ]), YAML_B3['options_traded_b3']['dt_input_format']
                                    )
                                )
                            )
                        )
                    #   closing price of spot and option
                    while \
                        (bl_break_spt_price == False) \
                        and (bl_break_opt_price == False):
                        for i, _ in enumerate(list_rows_mtm):
                            if i < len(list_rows_mtm) - 3:
                                #   option close
                                if list_rows_mtm[i][0] == YAML_B3['margens_teoricas_maximas_b3'][
                                    'key_instrumentos'] \
                                    and list_rows_mtm[i][4] != '0' \
                                    and list_rows_mtm[i][5] == dict_[
                                        YAML_B3['options_traded_b3']['keys']['ticker']]:
                                    dict_[YAML_B3['options_traded_b3']['keys']['close_opt']] = \
                                        float(list_rows_mtm[i][4].replace(',', '.'))
                                    bl_break_opt_price = True
                                    if bl_debug == True:
                                        print('OPT CLOSE: {}'.format(float(
                                            list_rows_mtm[i][4].replace(',', '.'))))
                                #   spot close
                                if list_rows_mtm[i][0] == YAML_B3['margens_teoricas_maximas_b3'][
                                    'key_instrumentos'] \
                                    and list_rows_mtm[i][4] != '0' \
                                    and list_rows_mtm[i][5] == dict_[YAML_B3['options_traded_b3'][
                                        'keys']['spot']]:
                                    dict_[YAML_B3['options_traded_b3']['keys']['close_spot']] = \
                                        float(list_rows_mtm[i][4].replace(',', '.'))
                                    bl_break_spt_price = True
                                    if bl_debug == True:
                                        print('SPOT CLOSE: {}'.format(float(
                                            list_rows_mtm[i][4].replace(',', '.'))))
                    #   risk-free rate
                    dict_[YAML_B3['options_traded_b3']['keys']['risk_free']] = r
                    #   days to maturity ratio
                    dict_[YAML_B3['options_traded_b3']['keys']['days_maturity_ratio']] = \
                        float(dict_[YAML_B3['options_traded_b3']['keys']['cd_exp']]) \
                        / float_days_year
                    #   implied volatility
                    if bl_debug == True:
                        print('*** INPUTS VOL IMP ***')
                        print(
                            float(dict_[YAML_B3['options_traded_b3']['keys']['close_spot']]), 
                            float(dict_[YAML_B3['options_traded_b3']['keys']['k']]), 
                            float(r), 
                            float(dict_[YAML_B3['options_traded_b3']['keys'][
                                'days_maturity_ratio']]), 
                            float(0.5), 
                            float(0), 
                            float(r),
                            float(dict_[YAML_B3['options_traded_b3']['keys']['close_opt']]), 
                            str(dict_[YAML_B3['options_traded_b3']['keys'][
                                'opt_type']].lower().replace('putt', 'put'))
                        )
                    try:
                        dict_[YAML_B3['options_traded_b3']['keys']['imp_vol']] = \
                            EuropeanOptions().implied_volatility(
                                float(dict_[YAML_B3['options_traded_b3']['keys']['close_spot']]), 
                                float(dict_[YAML_B3['options_traded_b3']['keys']['k']]), 
                                float(r), 
                                float(float(dict_[YAML_B3['options_traded_b3']['keys']['cd_exp']]) \
                                    / 365.0), 
                                float(0.5), 
                                float(0), 
                                float(r),
                                float(dict_[YAML_B3['options_traded_b3']['keys']['close_opt']]), 
                                str(dict_[YAML_B3['options_traded_b3']['keys']['opt_type']
                                        ].lower().replace('putt', 'put')
                                )
                            )
                    except KeyError:
                        dict_[YAML_B3['options_traded_b3']['keys']['imp_vol']] = float_value_error
                    #   delta
                    try:
                        dict_[YAML_B3['options_traded_b3']['keys']['delta']] = \
                            EuropeanOptions().delta(
                                float(dict_[YAML_B3['options_traded_b3']['keys']['close_spot']]), 
                                float(dict_[YAML_B3['options_traded_b3']['keys']['k']]), 
                                float(r), 
                                float(float(dict_[YAML_B3['options_traded_b3']['keys']['cd_exp']]) \
                                    / 365.0), 
                                dict_[YAML_B3['options_traded_b3']['keys']['imp_vol']], 
                                float(0), 
                                float(r),
                                str(dict_[YAML_B3['options_traded_b3']['keys']['opt_type']
                                    ].lower().replace('putt', 'put')
                                )
                            )
                    except KeyError:
                        dict_[YAML_B3['options_traded_b3']['keys']['delta']] = float_value_error
                    if bl_debug == True:
                        print('DELTA: {}'.format(dict_[YAML_B3['options_traded_b3']['keys'][
                            'delta']]))
                        print('IV: {}'.format(dict_[YAML_B3['options_traded_b3']['keys'][
                            'imp_vol']]))
                    #   appendig dictionary to list
                    list_ser.append(dict_)
                    if bl_debug == True:
                        print(dict_)
        # uploading list of dictionaries to pandas
        df_opt_in = pd.DataFrame(list_ser)
        # removing duplicates
        df_opt_in.drop_duplicates(inplace=True)
        # adding logging
        df_opt_in = DBLogs().audit_log(
            df_opt_in, 
            url,
            DatesBR().utc_from_dt(
                DatesBR().sub_working_days(DatesBR().curr_date, self.int_wd_bef)
            )
        )
        # returning dataframe
        return df_opt_in

    @property
    def carga_mtm_b3(self):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # url de exportação do arquivo de margens teóricas máximas b3
        url_mtm = YAML_B3['margens_teoricas_maximas_b3'][
            'url'].format(DatesBR().sub_working_days(DatesBR().curr_date,
                                                     self.int_wd_bef).strftime('%y%m%d'))
        # carga para a memória margens teóricas máximas b3
        zipfile = DirFilesManagement().get_zip_from_web_in_memory(
            url_mtm, bl_io_interpreting=YAML_B3[
                'margens_teoricas_maximas_b3']['bl_io_interpreting'],
            bl_verify=YAML_B3['margens_teoricas_maximas_b3']['bl_verify'])
        # retorno é um novo zip com um xlsx
        zipfile = ZipFile(zipfile)
        # coletando nome do arquivo txt contido dentro do zip
        list_zip_files = zipfile.namelist()
        file_name = list_zip_files.pop()
        # extraindo arquivo com fatores de risco primitivos b3 - txt
        name_file_extract = zipfile.open(file_name)
        # lendo csv com margens teóricas máximas b3
        list_rows_mtm = [
            str(str_).replace(r'\n', '').replace(
                "b'", '').replace("'", '').split(';')
            for str_ in name_file_extract.readlines() if len(
                str(str_).replace(r'\n', '').replace("b'", '').replace(
                    "'", '').split(';')) > 1
        ]
        list_rows_mtm = list_rows_mtm[1:]
        # retornando lista de linhas do csv corrente
        return list_rows_mtm

    @property
    def mtm_compra_venda(self, den_desagios=None,
                         key_stocks='stocks', key_funds='funds', key_etfs='etfs', key_bdrs='bdrs'):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS
        '''
        # setting variables
        list_num_desagios_compra = list()
        list_num_desagios_venda = list()
        list_exportacao = list()
        # carga de margens teóricas máximas crua
        list_rows_mtm = self.carga_mtm_b3
        # carga de ativos bovespa e bmf por tipo
        dict_ativos_bov_bmf_tipo = UP2DATAB3().security_category_name
        # selecionando apenas lista de ativos de interesse - ações, fiis, etfs, e bdrs
        list_ativos_bov_vista = HandlingLists().extend_lists(
            dict_ativos_bov_bmf_tipo[key_stocks], dict_ativos_bov_bmf_tipo[key_funds],
            dict_ativos_bov_bmf_tipo[key_etfs], dict_ativos_bov_bmf_tipo[key_bdrs])
        # removendo eventuais duplicatas
        list_ativos_bov_vista = HandlingLists().remove_duplicates(list_ativos_bov_vista)
        # alocando margens teóricas máximas em lista de dicionários
        for i, _ in enumerate(list_rows_mtm):
            if i < len(list_rows_mtm) - 3:
                #   setando variáveis de passagem
                dict_exportacao = dict()
                #   margem B3 requerida para passar posicionado no papel
                if list_rows_mtm[i][0] == YAML_B3['margens_teoricas_maximas_b3'][
                    'key_instrumentos'] \
                    and list_rows_mtm[i][4] != '0' \
                    and (len(list_rows_mtm[i][5]) == 5 or
                         len(list_rows_mtm[i][5]) == 6):
                    den_desagios = float(
                        list_rows_mtm[i][4].replace(',', '.'))
                #   numerador de preço com choque para cálculo de deságio por cenário
                #       dias úteis considerados: 1 a 10 dias úteis
                #       garantias exigidas calculadas de compra e venda
                for dias_uteis_choque in range(1, 11):
                    try:
                        if int(list_rows_mtm[i + dias_uteis_choque][1]) == \
                            dias_uteis_choque \
                                and list_rows_mtm[i + dias_uteis_choque][2] != '':
                            list_num_desagios_compra.append(float(
                                list_rows_mtm[i + dias_uteis_choque][2].replace(',', '.')))
                        if int(list_rows_mtm[i + dias_uteis_choque][
                            1]) == dias_uteis_choque and list_rows_mtm[
                                i + dias_uteis_choque][3] != '':
                            list_num_desagios_venda.append(float(
                                list_rows_mtm[i + dias_uteis_choque][3].replace(',', '.')))
                    except:
                        pass
                # executar o loop para listas vazias
                if (list_num_desagios_compra != []) and (list_num_desagios_venda != []):
                    # denominador com preço spot no fechamento para cálculo do deságio
                    #   B3 por cenário
                    if (list_rows_mtm[i][0] == YAML_B3['margens_teoricas_maximas_b3'][
                        'key_instrumentos']) and (list_rows_mtm[i][4] != '0') \
                            and all([type(x) == float for x in
                                     list_num_desagios_compra]) and (type(
                            den_desagios) == float):
                        # setando dicionário para o ticker de interesse
                        dict_exportacao[YAML_B3['margens_teoricas_maximas_b3'][
                            'col_incluir_ticker']] = list_rows_mtm[i][5]
                        # definindo preço de fechamento de referência
                        dict_exportacao['PRECO_FECHAMENTO'] = den_desagios
                        # definindo deságios para cenários de choque para o período de
                        #   interesse
                        for dias_uteis_choque in range(1, 11):
                            # deságio de compra/venda para ativo corrente negociado em mercado
                            #   secundário da b3
                            if list_rows_mtm[i][5] in list_ativos_bov_vista:
                                dict_exportacao[YAML_B3['margens_teoricas_maximas_b3'][
                                    'col_incluir_margem_b3_compra_d_mais_n'] + str(
                                    dias_uteis_choque)] = abs(den_desagios * (
                                        1 - list_num_desagios_compra[
                                            dias_uteis_choque - 1] / den_desagios))
                                dict_exportacao[YAML_B3['margens_teoricas_maximas_b3'][
                                    'col_incluir_desagio_b3_compra_d_mais_n'] + str(
                                    dias_uteis_choque)] = min([1, abs(1 - list_num_desagios_compra[
                                        dias_uteis_choque - 1] / den_desagios)])
                                dict_exportacao[YAML_B3['margens_teoricas_maximas_b3'][
                                    'col_incluir_margem_b3_venda_d_mais_n'] + str(
                                    dias_uteis_choque)] = \
                                    abs(den_desagios * (
                                        1 - list_num_desagios_venda[
                                            dias_uteis_choque - 1] / den_desagios))
                                dict_exportacao[YAML_B3['margens_teoricas_maximas_b3'][
                                    'col_incluir_desagio_b3_venda_d_mais_n'] + str(
                                    dias_uteis_choque)] = \
                                    min([1, abs(1 - list_num_desagios_venda[
                                        dias_uteis_choque - 1] / den_desagios)])
                            else:
                                dict_exportacao[YAML_B3['margens_teoricas_maximas_b3'][
                                    'col_incluir_margem_b3_compra_d_mais_n'] + str(
                                    dias_uteis_choque)] = abs(
                                    -list_num_desagios_compra[
                                        dias_uteis_choque - 1])
                                dict_exportacao[YAML_B3['margens_teoricas_maximas_b3'][
                                    'col_incluir_desagio_b3_compra_d_mais_n'] + str(
                                    dias_uteis_choque)] = \
                                    min([1, abs(list_num_desagios_compra[
                                        dias_uteis_choque - 1] / den_desagios - 1)])
                                dict_exportacao[YAML_B3['margens_teoricas_maximas_b3'][
                                    'col_incluir_margem_b3_venda_d_mais_n'] + str(
                                    dias_uteis_choque)] = \
                                    abs(-list_num_desagios_venda[dias_uteis_choque - 1])
                                dict_exportacao[YAML_B3['margens_teoricas_maximas_b3'][
                                    'col_incluir_desagio_b3_venda_d_mais_n'] + str(
                                    dias_uteis_choque)] = \
                                    min([1, abs(list_num_desagios_venda[
                                        dias_uteis_choque - 1] / den_desagios - 1)])
                    # appendando à lista de exportação
                    list_exportacao.append(dict_exportacao)
                    # ressetiando variável de passagem
                    list_num_desagios_compra = list()
                    list_num_desagios_venda = list()
        # appendando lista de exportação a um dataframe
        df_mtm_b3 = pd.DataFrame(list_exportacao)
        # drop not available rows
        df_mtm_b3.dropna(inplace=True)
        # sort tickers in alphabetic order
        df_mtm_b3.sort_values([YAML_B3['margens_teoricas_maximas_b3'][
            'col_incluir_ticker']], ascending=[True], inplace=True)
        # adding logging
        df_mtm_b3 = DBLogs().audit_log(
            df_mtm_b3, 
            YAML_B3['margens_teoricas_maximas_b3'][
                'url'].format(DatesBR().sub_working_days(
                    DatesBR().curr_date, self.int_wd_bef).strftime('%y%m%d')),
            DatesBR().utc_from_dt(
                DatesBR().sub_working_days(DatesBR().curr_date, self.int_wd_bef)
            )
        )
        # retornando deságios b3 de interesse
        return df_mtm_b3

    @property
    def fatores_primitivos_risco_b3(self):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # determinando url de download dos fatores primitivos de risco da b3
        url = YAML_B3['fatores_primitivos_risco_b3']['url'].format(DatesBR().sub_working_days(
                DatesBR().curr_date, self.int_wd_bef).strftime('%y%m%d'))
        # carga para a memória do arquivo em zip do anexo
        zipfile_fpr_b3 = DirFilesManagement().get_zip_from_web_in_memory(
            url,
            bl_io_interpreting=YAML_B3['fatores_primitivos_risco_b3']['bl_io_interpreting'],
            bl_verify=YAML_B3['fatores_primitivos_risco_b3']['bl_verify'])
        # retorno é um novo zip com um txt quanto aos fatores de risco primitivos da b3, com isso
        #   desziparesse arquivo
        zipfile_fpr_b3_novo = ZipFile(zipfile_fpr_b3)
        # coletando nome do arquivo txt contido dentro do zip
        list_zip_files = zipfile_fpr_b3_novo.namelist()
        file_name = list_zip_files.pop()
        # extraindo arquivo com fatores de risco primitivos b3 - txt
        extracted_file_fpr_b3 = zipfile_fpr_b3_novo.open(file_name)
        # coletando em memória fatores de risco primitivos b3 - importando em dataframe e tratando
        #   como um csv, separando arquivo com base no caractére de quebra das instâncias (;)
        reader = pd.read_csv(extracted_file_fpr_b3, sep=';', skiprows=1, header=None, names=YAML_B3[
            'fatores_primitivos_risco_b3']['cols_fatores_primitivos_b3'])
        df_fpr_b3 = pd.DataFrame(reader)
        # substituindo valores NA por valor padrão
        df_fpr_b3.fillna(YAML_B3['fatores_primitivos_risco_b3']['valor_na_substituir'],
                         inplace=True)
        # alterando tipo de colunas de interesse
        df_fpr_b3 = df_fpr_b3.astype({
            YAML_B3['fatores_primitivos_risco_b3']['cols_fatores_primitivos_b3'][0]: int,
            YAML_B3['fatores_primitivos_risco_b3']['cols_fatores_primitivos_b3'][1]: int,
            YAML_B3['fatores_primitivos_risco_b3']['cols_fatores_primitivos_b3'][2]: str,
            YAML_B3['fatores_primitivos_risco_b3']['cols_fatores_primitivos_b3'][3]: int,
            YAML_B3['fatores_primitivos_risco_b3']['cols_fatores_primitivos_b3'][4]: int,
            YAML_B3['fatores_primitivos_risco_b3']['cols_fatores_primitivos_b3'][5]: str,
            YAML_B3['fatores_primitivos_risco_b3']['cols_fatores_primitivos_b3'][6]: str,
            YAML_B3['fatores_primitivos_risco_b3']['cols_fatores_primitivos_b3'][7]: str,
            YAML_B3['fatores_primitivos_risco_b3']['cols_fatores_primitivos_b3'][8]: int,
            YAML_B3['fatores_primitivos_risco_b3']['cols_fatores_primitivos_b3'][9]: int,
            YAML_B3['fatores_primitivos_risco_b3']['cols_fatores_primitivos_b3'][10]: int,
        })
        # incluindo coluna do nome do formato da variação
        df_fpr_b3[YAML_B3['fatores_primitivos_risco_b3']['col_incluir_nome_formato_variacao']] = \
            [str(YAML_B3['fatores_primitivos_risco_b3']['variacao_fatores_primitivos']['tipo'][
                int_formato_variacao]) for int_formato_variacao in df_fpr_b3[YAML_B3[
                    'fatores_primitivos_risco_b3']['cols_fatores_primitivos_b3'][3]]]
        # incluindo coluna de origem do instrumento
        df_fpr_b3[YAML_B3['fatores_primitivos_risco_b3']['col_incluir_nome_origem_instrumento']] = \
            [str(YAML_B3['fatores_primitivos_risco_b3']['variacao_fatores_primitivos'][
                'origem_instrumento'][int(int_origem_instrumento)]) for int_origem_instrumento in
                df_fpr_b3[YAML_B3['fatores_primitivos_risco_b3']['cols_fatores_primitivos_b3'][7]]]
        # incluindo coluna de origem do instrumento
        df_fpr_b3[YAML_B3['fatores_primitivos_risco_b3']['col_incluir_criterio_capitalizacao']] = \
            [str(YAML_B3['fatores_primitivos_risco_b3']['variacao_fatores_primitivos'][
                'criterio_capitalizacao'][int(int_criterio_cap)]) for int_criterio_cap in
                df_fpr_b3[YAML_B3['fatores_primitivos_risco_b3']['cols_fatores_primitivos_b3'][10]]]
        # adding logging
        df_fpr_b3 = DBLogs().audit_log(
            df_fpr_b3, 
            url,
            DatesBR().utc_from_dt(
                DatesBR().sub_working_days(DatesBR().curr_date, self.int_wd_bef)
            )
        )
        # retornando dataframe com fatores primitivos de risco b3
        return df_fpr_b3

    @property
    def collateral_acc_spot_bov_b3(self, dict_repalce_str_erros={
            'ETF': ''}, col_ticker='TICKER', col_isin='ISIN', col_limite_qtds='LIMITE_QUANTIDADES',
            col_mes_arquivo='MES_ARQUIVO', key_limite_qtds='Limite (quantidade)'):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # setting variables
        list_exportacao = list()
        # importing to memory html, in order to catch the url of the collateral accepted by b3, 
        #   regarding spot bov
        html_parser_acoes_units_etfs_aceitos_garantia_b3 = HtmlHndler().html_lxml_parser(
            YAML_B3['collateral_acc_spot_bov_b3']['url'], bl_verify=False)
        url_acoes_units_etfs_aceitos_garantia_b3 = HtmlHndler().html_lxml_xpath(
            html_parser_acoes_units_etfs_aceitos_garantia_b3,
            YAML_B3['collateral_acc_spot_bov_b3'][
                'xpath_url_acoes_units_etfs_aceitos_garantia_b3'])[0]
        url_acoes_units_etfs_aceitos_garantia_b3 = StrHandler().get_string_after_substr(
            url_acoes_units_etfs_aceitos_garantia_b3, YAML_B3['collateral_acc_spot_bov_b3'][
            'data_files_collateral_acc_spot_bov_b3'])
        # creating complete url
        url_acoes_units_etfs_aceitos_garantia_b3 = \
            YAML_B3['collateral_acc_spot_bov_b3']['url_raiz_documentos_b3'] \
            + YAML_B3['collateral_acc_spot_bov_b3'][
            'data_files_collateral_acc_spot_bov_b3'] + \
            url_acoes_units_etfs_aceitos_garantia_b3
        # openning zip in memory
        list_unzipped_files = DirFilesManagement().get_zip_from_web_in_memory(
            url_acoes_units_etfs_aceitos_garantia_b3, bl_verify=False)
        if str(type(list_unzipped_files)) != "<class 'list'>":
            list_unzipped_files = [list_unzipped_files]
        # looping within files in zip
        for file_unz in list_unzipped_files[1:]:
            #   creating columns month-file
            str_mes_arquivo_garantias_b3_corrente = StrHandler().find_between(
                str(file_unz), '_', '.xlsx').replace(' ', '').lower()
            #   looping within sheet names
            for nome_plan_ativa in [
                StrHandler().latin_characters(
                    YAML_B3['collateral_acc_spot_bov_b3']['plan_acoes_bdrs_etfs_units']), 
                YAML_B3['collateral_acc_spot_bov_b3']['plan_adrs']
            ]:
                #   defining column ticker
                if nome_plan_ativa == 'ADR':
                    col_cod_ticker = 'Código'
                else:
                    col_cod_ticker = 'Código'
                #   openning dataframe in memory
                reader = pd.read_excel(
                    file_unz, sheet_name=nome_plan_ativa)
                df_coll_acc_spot_bov = pd.DataFrame(reader)
                #   creating serialized list with spot bov collateral accepted
                for _, row in df_coll_acc_spot_bov.iterrows():
                    list_exportacao.append({
                        col_mes_arquivo: str_mes_arquivo_garantias_b3_corrente.upper(),
                        col_ticker: row[col_cod_ticker],
                        col_isin: row[col_isin],
                        col_limite_qtds: row[key_limite_qtds]
                    })
        # creating dataframe
        df_coll_acc_spot_bov = pd.DataFrame(list_exportacao)
        # removing duplicates
        df_coll_acc_spot_bov[col_ticker] = [StrHandler().replace_all(
            str(x), dict_repalce_str_erros) for x in df_coll_acc_spot_bov[
                col_ticker].tolist()]
        # adding logging
        df_fpr_b3 = DBLogs().audit_log(
            df_fpr_b3, 
            url_acoes_units_etfs_aceitos_garantia_b3,
            DatesBR().utc_from_dt(DatesBR().curr_date)
        )
        # returning dataframe
        return df_coll_acc_spot_bov

    def cenarios_tipo_curva(self, list_cenarios_tipo_curva=None):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # setting variables
        list_ser_exportacao_curvas = list()
        # validando tipo de variáveis de interesse
        if list_cenarios_tipo_curva != None:
            #   convertendo para lista, caso não seja
            if type(list_cenarios_tipo_curva) != list:
                list_cenarios_tipo_curva = [list_cenarios_tipo_curva]
            #   alterando tipo de instâncias
            list_cenarios_tipo_curva = [str(x)
                                        for x in list_cenarios_tipo_curva]
        # definindo url com cenários de risco - tipo de curva
        url_cenarios_tipo_curva = YAML_B3['cenarios_risco_tipo_curva']['tipo_curva']['url'].format(
            DatesBR().sub_working_days(DatesBR().curr_date, YAML_B3['cenarios_risco_tipo_curva'][
                'tipo_curva']['self.int_wd_bef']).strftime('%y%m%d'))
        # baixando em memória zip com cenário de tipo curva da b3
        list_txts_tipos_curva_cenarios = DirFilesManagement().get_zip_from_web_in_memory(
            url_cenarios_tipo_curva, bl_verify=YAML_B3['cenarios_risco_tipo_curva']['bl_verify'],
            bl_io_interpreting=YAML_B3['cenarios_risco_tipo_curva']['bl_io_interpreting'],
            timeout=YAML_B3['cenarios_risco_tipo_curva']['tipo_curva']['timeout'])
        # loopando em torno dos arquivos em txt, abrindo em memória no dataframe pandas
        for extracted_file_tipo_curva_cenario in list_txts_tipos_curva_cenarios:
            #   caso a variável list_cenarios_tipo_curva seja diferente de none, o usuário tem
            #       interesse que apenas uma parte dos cenários seja considerado, com isso, caso
            #       o cenário corrente não esteja nessa amostra passar para o próximo
            if (list_cenarios_tipo_curva != None) and (all([StrHandler().find_substr_str(
                    extracted_file_tipo_curva_cenario.name, substr) == False for substr in
                    list_cenarios_tipo_curva])):
                continue
            #   abrindo em memória arquivo corrente - pandas dataframe
            reader = pd.read_csv(extracted_file_tipo_curva_cenario, sep=YAML_B3[
                'cenarios_risco_tipo_curva']['sep_instancias'], skiprows=1, header=None,
                names=YAML_B3['cenarios_risco_tipo_curva']['cols_cenarios_tipo_curvas'], decimal=',')
            df_cenarios_passagem = pd.DataFrame(reader)
            #   preenchendo com valor padrão campos com valor indisponível
            df_cenarios_passagem.fillna(YAML_B3['cenarios_risco_tipo_curva']['fill_na_padrao'],
                                        inplace=True)
            #   alterando para dicionário dataframe corrente, visando consolidar em uma lista e
            #       importar para uma dataframe para exportação
            list_passagem = df_cenarios_passagem.to_dict(orient=YAML_B3[
                'cenarios_risco_tipo_curva']['orient_df'])
            list_ser_exportacao_curvas.extend(list_passagem)
        # adicionando lista a um dataframe a ser exportado
        df_cenarios_tipo_curva = pd.DataFrame.from_dict(
            list_ser_exportacao_curvas)
        # alterando tipo de variáveis de colunas de interesse
        df_cenarios_tipo_curva = df_cenarios_tipo_curva.astype({
            YAML_B3['cenarios_risco_tipo_curva']['cols_cenarios_tipo_curvas'][0]: int,
            YAML_B3['cenarios_risco_tipo_curva']['cols_cenarios_tipo_curvas'][1]: int,
            YAML_B3['cenarios_risco_tipo_curva']['cols_cenarios_tipo_curvas'][2]: int,
            YAML_B3['cenarios_risco_tipo_curva']['cols_cenarios_tipo_curvas'][3]: int,
            YAML_B3['cenarios_risco_tipo_curva']['cols_cenarios_tipo_curvas'][4]: int,
            YAML_B3['cenarios_risco_tipo_curva']['cols_cenarios_tipo_curvas'][5]: int,
            YAML_B3['cenarios_risco_tipo_curva']['cols_cenarios_tipo_curvas'][6]: int,
            YAML_B3['cenarios_risco_tipo_curva']['cols_cenarios_tipo_curvas'][7]: float,
            YAML_B3['cenarios_risco_tipo_curva']['cols_cenarios_tipo_curvas'][8]: float,
        })
        # adding logging
        df_cenarios_tipo_curva = DBLogs().audit_log(
            df_cenarios_tipo_curva, 
            url_cenarios_tipo_curva,
            DatesBR().utc_from_dt(DatesBR().curr_date)
        )
        # exportando dataframe de interesse
        return df_cenarios_tipo_curva

    def cenarios_risco_tipo_spot(self, list_cenarios_tipo_spot=None):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # setting variables
        list_ser_exportacao_curvas = list()
        # validando tipo de variáveis de interesse
        if list_cenarios_tipo_spot != None:
            #   convertendo para lista, caso não seja
            if type(list_cenarios_tipo_spot) != list:
                list_cenarios_tipo_spot = [list_cenarios_tipo_spot]
            #   alterando tipo de instâncias
            list_cenarios_tipo_spot = [str(x) for x in list_cenarios_tipo_spot]
        # definindo url com cenários de risco - tipo de curva
        url_cenarios_tipo_spot = YAML_B3['cenarios_risco_tipo_spot']['tipo_curva']['url'].format(
            DatesBR().sub_working_days(DatesBR().curr_date, YAML_B3['cenarios_risco_tipo_spot'][
                'tipo_curva']['self.int_wd_bef']).strftime('%y%m%d'))
        # baixando em memória zip com cenário de tipo curva da b3
        list_txts_tipos_spot_cenarios = DirFilesManagement().get_zip_from_web_in_memory(
            url_cenarios_tipo_spot, bl_verify=YAML_B3['cenarios_risco_tipo_spot']['bl_verify'],
            bl_io_interpreting=YAML_B3['cenarios_risco_tipo_spot']['bl_io_interpreting'],
            timeout=YAML_B3['cenarios_risco_tipo_spot']['tipo_curva']['timeout'])
        # loopando em torno dos arquivos em txt, abrindo em memória no dataframe pandas
        for extracted_file_tipo_curva_cenario in list_txts_tipos_spot_cenarios:
            #   caso a variável list_cenarios_tipo_spot seja diferente de none, o usuário tem
            #       interesse que apenas uma parte dos cenários seja considerado, com isso, caso o
            #       cenário corrente não esteja nessa amostra passar para o próximo
            if (list_cenarios_tipo_spot != None) and (all([StrHandler().find_substr_str(
                    extracted_file_tipo_curva_cenario.name, substr) == False for substr in
                    list_cenarios_tipo_spot])):
                continue
            #   abrindo em memória arquivo corrente - pandas dataframe
            reader = pd.read_csv(extracted_file_tipo_curva_cenario, sep=YAML_B3[
                'cenarios_risco_tipo_spot']['sep_instancias'], skiprows=1, header=None,
                names=YAML_B3['cenarios_risco_tipo_spot']['cols_cenarios_tipo_curvas'], decimal=',')
            df_cenarios_passagem = pd.DataFrame(reader)
            #   preenchendo com valor padrão campos com valor indisponível
            df_cenarios_passagem.fillna(
                YAML_B3['cenarios_risco_tipo_spot']['fill_na_padrao'], inplace=True)
            #   alterando para dicionário dataframe corrente, visando consolidar em uma lista e
            #       importar para uma dataframe para exportação
            list_passagem = df_cenarios_passagem.to_dict(orient=YAML_B3[
                'cenarios_risco_tipo_spot']['orient_df'])
            list_ser_exportacao_curvas.extend(list_passagem)
        # adicionando lista a um dataframe a ser exportado
        df_cenarios_tipo_spot = pd.DataFrame.from_dict(
            list_ser_exportacao_curvas)
        # alterando tipo de variáveis de colunas de interesse
        df_cenarios_tipo_spot = df_cenarios_tipo_spot.astype({
            YAML_B3['cenarios_risco_tipo_spot']['cols_cenarios_tipo_curvas'][0]: int,
            YAML_B3['cenarios_risco_tipo_spot']['cols_cenarios_tipo_curvas'][1]: int,
            YAML_B3['cenarios_risco_tipo_spot']['cols_cenarios_tipo_curvas'][2]: int,
            YAML_B3['cenarios_risco_tipo_spot']['cols_cenarios_tipo_curvas'][3]: int,
            YAML_B3['cenarios_risco_tipo_spot']['cols_cenarios_tipo_curvas'][4]: int,
            YAML_B3['cenarios_risco_tipo_spot']['cols_cenarios_tipo_curvas'][5]: float,
            YAML_B3['cenarios_risco_tipo_spot']['cols_cenarios_tipo_curvas'][6]: float,
        })
        # adicionando colunas de interesse - nome tipo do cenário
        df_cenarios_tipo_spot[YAML_B3['cenarios_risco_tipo_spot']['col_criar_nome_tipo_cenario']] = \
            [YAML_B3['cenarios_risco_tipo_spot']['variaveis_tipo_cenarios_curvas'][x] for x in
                df_cenarios_tipo_spot[YAML_B3['cenarios_risco_tipo_spot'][
                    'cols_cenarios_tipo_curvas'][3]].tolist()]
        # adding logging
        df_cenarios_tipo_spot = DBLogs().audit_log(
            df_cenarios_tipo_spot, 
            url_cenarios_tipo_spot,
            DatesBR().utc_from_dt(DatesBR().curr_date)
        )
        # exportando dataframe de interesse
        return df_cenarios_tipo_spot

    def fprs_cenarios_curva_spot_merge(self,
                                       list_cenarios_tipo_curva=['2962', '2906', '2945', '2924',
                                                                 '2944', '2925', '2946',
                                                                 '2901', '2902', '2934'],
                                       list_cenarios_tipo_spot=['2962', '2906', '2945', '2924',
                                                                '2944', '2925', '2946',
                                                                '2901', '2902', '2934'],
                                       int_cenario_sup_avaliacao=529, bl_debug=True):
        '''
        DOCSTRING: UNINDO FATORES PRIMITIVOS DE RISCO, CURVAS SPOT E POR TIPO EM UM DATAFRAME COM 
            OS IDS DE CENÁRIOS DE ITNERESSE
        INPUTS: -
        OUTPUTS: DATAFRAME
        '''
        # criando dataframe de tipos de frps de interesse
        df_tipos_fprs = pd.DataFrame(
            YAML_B3['fatores_primitivos_risco_b3']['tipos_fprs'])
        # convertendo tipo de colunas de interesse
        df_tipos_fprs = df_tipos_fprs.astype({
            list(YAML_B3['fatores_primitivos_risco_b3']['tipos_fprs'].keys())[0]: int,
            list(YAML_B3['fatores_primitivos_risco_b3']['tipos_fprs'].keys())[1]: str,
        })
        if bl_debug == True:
            print('*** TIPO FRPS ***')
            print(df_tipos_fprs)
        # * arquivos de pregão b3 - cenários tipo curva
        df_cenarios_tipo_curva = self.cenarios_tipo_curva(list_cenarios_tipo_curva)
        # removendo colunas de interesse
        df_cenarios_tipo_curva.drop([
            YAML_B3['cenarios_risco_tipo_curva']['col_tipo'],
            YAML_B3['cenarios_risco_tipo_curva']['col_phi_2']
        ], axis=1, inplace=True)
        # limitando cenários de tipos de curvas dentro do range de interesse
        df_cenarios_tipo_curva = df_cenarios_tipo_curva[df_cenarios_tipo_curva[YAML_B3[
            'cenarios_risco_tipo_curva']['col_id_cenario']].isin(range(
                YAML_B3['cenarios_risco_tipo_curva']['int_inf_range_cenarios'],
                YAML_B3['cenarios_risco_tipo_curva']['int_sup_range_cenarios']))].reset_index(
                    drop=True)
        # eliminando cenários prospectivos
        df_cenarios_tipo_curva = df_cenarios_tipo_curva[df_cenarios_tipo_curva['ID_CENARIO'].isin(
            range(1, int_cenario_sup_avaliacao))].reset_index(drop=True)
        if bl_debug == True:
            print('*** TIPO CURVA ***')
            print(df_cenarios_tipo_curva.info())
            print(df_cenarios_tipo_curva)
        # * arquivos de pregão b3 - cenários risco tipo spot
        df_cenarios_tipo_spot = self.cenarios_risco_tipo_spot(list_cenarios_tipo_spot)
        # removendo colunas de interesse
        df_cenarios_tipo_spot.drop([
            YAML_B3['cenarios_risco_tipo_curva']['col_tipo'],
            YAML_B3['cenarios_risco_tipo_curva']['col_phi_2']
        ], axis=1, inplace=True)
        # limitando cenários de curvas spot
        df_cenarios_tipo_spot = df_cenarios_tipo_spot[df_cenarios_tipo_spot[YAML_B3[
            'cenarios_risco_tipo_spot']['col_id_cenario']].isin(range(
                YAML_B3['cenarios_risco_tipo_spot']['int_inf_range_cenarios'],
                YAML_B3['cenarios_risco_tipo_spot']['int_sup_range_cenarios']))].reset_index(
                    drop=True)
        # eliminando cenários prospectivos
        df_cenarios_tipo_spot = df_cenarios_tipo_spot[df_cenarios_tipo_spot['ID_CENARIO'].isin(
            range(1, int_cenario_sup_avaliacao))].reset_index(drop=True)
        if bl_debug == True:
            print('*** TIPO SPOT ***')
            print(df_cenarios_tipo_spot.info())
            print(df_cenarios_tipo_spot)
        # * arquivos de pregão b3 - fatores primitivos de risco
        df_fpr_b3 = self.fatores_primitivos_risco_b3
        # removendo colunas de interesse
        df_fpr_b3.drop([
            YAML_B3['fatores_primitivos_risco_b3']['col_tipo']
        ], axis=1, inplace=True)
        # limitando fprs de interesse para fins de cálculo de swaps
        df_fpr_b3 = df_fpr_b3[df_fpr_b3[
            YAML_B3['fatores_primitivos_risco_b3']['col_id_fpr']].isin(list(YAML_B3[
                'fatores_primitivos_risco_b3']['dict_fprs'].keys()))]
        if bl_debug == True:
            print('*** TIPO FPR ***')
            print(df_fpr_b3.info())
            print(df_fpr_b3)
        # * início tratamento de dados
        # inner-join fprs com dataframe de tipos de frps
        list_colunas_frps = df_fpr_b3.columns
        if bl_debug == True:
            print('*** COLUNAS FRPS ***')
            print(list_colunas_frps)
        list_colunas_frp = ['ID_FPR', 'ID_CENARIO', 'TIPO_FPR'] + list_colunas_frps.drop([
            'ID_FPR', 'NOME_FPR']).to_list()
        if bl_debug == True:
            print('*** TIPO FPR B3 ***')
            print(df_fpr_b3)
            print('*** TIPOS FPRS ***')
            print(df_tipos_fprs)
        df_fpr_b3 = df_fpr_b3.merge(
            df_tipos_fprs, how='inner', left_on='ID_FPR', right_on='ID_FPR')
        df_fpr_b3 = df_fpr_b3.reindex(columns=list_colunas_frp)
        # concat cenários tipo curva e spot
        df_cenarios_tipo_curva = pd.concat([df_cenarios_tipo_curva, df_cenarios_tipo_spot],
                                           ignore_index=True, sort=False)
        if bl_debug == True:
            print('*** CENÁRIOS TIPO CURVA & SPOT ***')
            print(df_cenarios_tipo_curva)
        # inner-join fprs com cenarios spot e tipo curvas
        df_fpr_b3 = df_fpr_b3.merge(df_cenarios_tipo_curva, how='inner', left_on='ID_FPR',
                                    right_on='ID_FPR')
        # preenchendo coluna nome fpr
        df_fpr_b3['NOME_FPR'] = [YAML_B3['fatores_primitivos_risco_b3']['dict_fprs'][x]
                                 for x in df_fpr_b3['ID_FPR']]
        # alterando tipo de colunas de interesse
        df_fpr_b3 = df_fpr_b3.astype({
            'DIAS_CORRIDOS_VERTICE': 'Int64',
            'DIAS_HOLDING_PERIOD': 'Int64',
            'DIAS_SAQUE_VERTICE': 'Int64',
        })
        # limitando colunas de interesse
        df_fpr_b3 = df_fpr_b3[[
            'ID_FPR',
            'NOME_FPR',
            'TIPO_FPR',
            'FORMATO_VARIACAO',
            'ID_GRUPO_FPR',
            'ID_CAMARA_INDICADOR',
            'ID_INSTRUMENTO_INDICADOR',
            'ORIGEM_INSTURMENTO_INDICADOR',
            'BASE',
            'BASE_INTERPOLACAO',
            'CRITERIO_CAPITALIZACAO',
            'ID_CENARIO_y',
            'INT_TIPO_CENARIO',
            'DIAS_CORRIDOS_VERTICE',
            'DIAS_HOLDING_PERIOD',
            'DIAS_SAQUE_VERTICE',
            'VALOR_PHI_1',
        ]]
        # alterando nome de colunas de interesse
        df_fpr_b3.rename(columns={
            'ID_CENARIO_y': 'ID_CENARIO'
        }, inplace=True)
        if bl_debug == True:
            print('*** DF FPRS CONSOLIDADO ***')
            print(df_fpr_b3)
        # adding logging
        df_fpr_b3 = DBLogs().audit_log(
            df_fpr_b3, 
            None,
            DatesBR().utc_from_dt(DatesBR().curr_date)
        )
        # retornando fpr completo
        return df_fpr_b3

    def cenarios_risco_tipo_spot(self, list_cenarios_tipo_spot=None):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # setting variables
        list_ser_exportacao_curvas = list()
        # validando tipo de variáveis de interesse
        if list_cenarios_tipo_spot != None:
            #   convertendo para lista, caso não seja
            if type(list_cenarios_tipo_spot) != list:
                list_cenarios_tipo_spot = [list_cenarios_tipo_spot]
            #   alterando tipo de instâncias
            list_cenarios_tipo_spot = [str(x) for x in list_cenarios_tipo_spot]
        try:
            # definindo url com cenários de risco - tipo de curva
            url_cenarios_tipo_spot = YAML_B3['cenarios_risco_tipo_spot']['tipo_curva'][
                'url'].format(DatesBR().sub_working_days(
                    DatesBR().curr_date, YAML_B3['cenarios_risco_tipo_spot'][
                        'tipo_curva']['self.int_wd_beferiores']).strftime('%y%m%d'))
            # baixando em memória zip com cenário de tipo curva da b3
            list_txts_tipos_spot_cenarios = DirFilesManagement().get_zip_from_web_in_memory(
                url_cenarios_tipo_spot, bl_verify=YAML_B3['cenarios_risco_tipo_spot']['bl_verify'], 
                bl_io_interpreting=YAML_B3['cenarios_risco_tipo_spot']['bl_io_interpreting'], 
                timeout=YAML_B3['cenarios_risco_tipo_spot']['tipo_curva']['timeout'])
        except:
            url_cenarios_tipo_spot = YAML_B3['cenarios_risco_tipo_spot'][
                'tipo_curva']['url'].format(
            DatesBR().sub_working_days(DatesBR().curr_date, (YAML_B3['cenarios_risco_tipo_spot'][
                'tipo_curva']['self.int_wd_beferiores'] + 1)).strftime('%y%m%d'))
            list_txts_tipos_spot_cenarios = DirFilesManagement().get_zip_from_web_in_memory(
                url_cenarios_tipo_spot, bl_verify=YAML_B3['cenarios_risco_tipo_spot']['bl_verify'], 
                bl_io_interpreting=YAML_B3['cenarios_risco_tipo_spot']['bl_io_interpreting'], 
                timeout=YAML_B3['cenarios_risco_tipo_spot']['tipo_curva']['timeout'])
        # loopando em torno dos arquivos em txt, abrindo em memória no dataframe pandas
        for extracted_file_tipo_curva_cenario in list_txts_tipos_spot_cenarios:
            #   caso a variável list_cenarios_tipo_spot seja diferente de none, o usuário tem 
            #       interesse que apenas uma parte dos cenários seja considerado, com isso, caso 
            #       o cenário corrente não esteja nessa amostra passar para o próximo
            if (list_cenarios_tipo_spot != None) and (all([StrHandler().find_substr_str(
                extracted_file_tipo_curva_cenario.name, substr) == False for substr in 
                list_cenarios_tipo_spot])): continue
            #   abrindo em memória arquivo corrente - pandas dataframe
            reader = pd.read_csv(extracted_file_tipo_curva_cenario, sep=YAML_B3[
                'cenarios_risco_tipo_spot'][
                'sep_instancias'], skiprows=1, header=None, names=YAML_B3[
                    'cenarios_risco_tipo_spot']['cols_cenarios_tipo_curvas'], decimal=',')
            df_cenarios_passagem = pd.DataFrame(reader)
            #   preenchendo com valor padrão campos com valor indisponível
            df_cenarios_passagem.fillna(YAML_B3['cenarios_risco_tipo_spot']['fill_na_padrao'], 
                                        inplace=True)
            #   alterando para dicionário dataframe corrente, visando consolidar em uma lista
            #       e importar para uma dataframe para exportação
            list_passagem = df_cenarios_passagem.to_dict(orient=YAML_B3[
                'cenarios_risco_tipo_spot']['orient_df'])
            list_ser_exportacao_curvas.extend(list_passagem)
        # adicionando lista a um dataframe a ser exportado
        df_cenarios_tipo_spot = pd.DataFrame.from_dict(list_ser_exportacao_curvas)
        # alterando tipo de variáveis de colunas de interesse
        df_cenarios_tipo_spot = df_cenarios_tipo_spot.astype({
            YAML_B3['cenarios_risco_tipo_spot']['cols_cenarios_tipo_curvas'][0]: int,
            YAML_B3['cenarios_risco_tipo_spot']['cols_cenarios_tipo_curvas'][1]: int,
            YAML_B3['cenarios_risco_tipo_spot']['cols_cenarios_tipo_curvas'][2]: int,
            YAML_B3['cenarios_risco_tipo_spot']['cols_cenarios_tipo_curvas'][3]: int,
            YAML_B3['cenarios_risco_tipo_spot']['cols_cenarios_tipo_curvas'][4]: int,
            YAML_B3['cenarios_risco_tipo_spot']['cols_cenarios_tipo_curvas'][5]: float,
            YAML_B3['cenarios_risco_tipo_spot']['cols_cenarios_tipo_curvas'][6]: float,
        })
        # adicionando colunas de interesse - nome tipo do cenário
        df_cenarios_tipo_spot[YAML_B3['cenarios_risco_tipo_spot'][
            'col_criar_nome_tipo_cenario']] = \
            [YAML_B3['cenarios_risco_tipo_spot']['variaveis_tipo_cenarios_curvas'][x] for x in 
                df_cenarios_tipo_spot[YAML_B3['cenarios_risco_tipo_spot'][
                    'cols_cenarios_tipo_curvas'][3]].tolist()]
        # adding logging
        df_cenarios_tipo_spot = DBLogs().audit_log(
            df_cenarios_tipo_spot, 
            url_cenarios_tipo_spot,
            DatesBR().utc_from_dt(DatesBR().curr_date)
        )
        # exportando dataframe de interesse
        return df_cenarios_tipo_spot

    @property
    def trading_report(self, 
        el_pric_rpt='PricRpt', el_trad_dt='TradDt', el_dt='Dt', el_scty_id='SctyId', 
        el_ticker='TckrSymb', el_fin_ins_id='FinInstrmId', el_othr_id='OthrId', el_id='Id', 
        el_tp='Tp', el_pr_try='Prtry', el_plc_of_listg='PlcOfListg', el_mkt_id_rcd='MktIdrCd', 
        el_trad_dtls='TradDtls', el_trad_qty='TradQty', el_fin_att='FinInstrmAttrbts', 
        el_mkt_data_strm_id='MktDataStrmId', el_nt_fin='NtlFinVol', el_int_fin='IntlFinVol', 
        el_opn_instr='OpnIntrst', el_fin_instrm_qt='FinInstrmQty', el_best_ask='BestAskPric',
        el_first_pric='FrstPric', el_min_pric='MinPric', el_max_prec='MaxPric', 
        el_avg_pric='TradAvrgPric', el_last_pric='LastPric', el_rglr_qt='RglrTxsQty', 
        el_rglr_trad='RglrTraddCtrcts', el_ntl_rglr_vol='NtlRglrVol', el_intl_rglr_vol='IntlRglrVol', 
        el_oscn_pctg='OscnPctg', col_dt_transacao='DT_TRANSACAO', col_ticker='TICKER', col_id='ID', 
        col_id_prtry='ID_PRTRY', col_mercado='MERCADO', col_qtd_neg='QTD_NEGOCIOS', 
        col_str_exercicio='STR_EXERCICIO', col_notional_neg_dom='NOTIONAL_NEG_MOEDA_DOM', 
        col_notional_neg_int='NOTIONAL_NEG_MOEDA_INT', 
        col_qtd_contratos_aberto='QTD_CONTRATOS_ABERTO', col_qtd_contratos_neg='QTD_CONTRATOS_NEG',
        col_melhor_ask='MELHOR_ASK', col_prim_prec='PRIMEIRO_PRECO', col_menor_preco='MENOR_PRECO', 
        col_maior_prec='MAIOR_PRECO', col_preco_med='PRECO_MEDIO', col_ult_preco='ULT_PRECO', 
        col_qtd_negocios_sess_reg='QTD_NEGOCIOS_SESS_REG', 
        col_qtd_negociada_sess_reg='QTD_NEGOCIADA_SESS_REG', 
        col_notional_neg_dom_sess_reg='NOTIONAL_NEG_MOEDA_DOM_SESS_REG', 
        col_notional_neg_int_sess_reg='NOTIONAL_NEG_MOEDA_INT_SESS_REG', 
        col_pct_osc_int='PCT_OSC_INT', formato_dt='AAAA-MM-DD', qtd_erro=-1, 
        col_ticker_cd='Symbol', col_tipo_mercado_cd='TIPO_MERCADO'):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # setting variables
        list_ser = list()
        # instruments and its respective market
        df_ = self.tradable_securities(self.int_wd_bef)
        # url to download maximum theoretical margins
        url = YAML_B3['trading_report']['url'].format(DatesBR().sub_working_days(
            DatesBR().curr_date, self.int_wd_bef).strftime('%y%m%d')
        )
        # import to memory trading report zip
        zipfile = DirFilesManagement().get_zip_from_web_in_memory(
            url, 
            bl_io_interpreting=YAML_B3['trading_report']['bl_io_interpreting'], 
            bl_verify=YAML_B3['trading_report']['bl_verify'])
        # retorno é um novo zip com um txt quanto aos fatores de risco primitivos da b3, 
        #   com isso deszipar esse arquivo 
        zipfile = ZipFile(zipfile)
        # collecting file name within zip
        list_zip_f = zipfile.namelist()
        # loopando ao longo dos arquivos de interesse
        for file_xml in list_zip_f:
            #   extraindo arquivo com boletim de negociação b3
            extracted_file_b3 = zipfile.open(file_xml)
            #   coletando em memória bvbg do boletim de negociação
            soup_xml_bn = XMLFiles().xml_memory_parser(extracted_file_b3)
            #   loopando em torno do elemento PricRpt (repositório de preços e volume de negociação) 
            #       e coletando informações de interesse
            for content_xml_bn in soup_xml_bn.find_all(el_pric_rpt):
                #   definindo dicionário de passagem
                dict_ = dict()
                #   data de transação 
                dict_[col_dt_transacao] = content_xml_bn.find(el_trad_dt).find(el_dt).get_text()
                #   ticker
                dict_[col_ticker] = content_xml_bn.find(el_scty_id).find(el_ticker).get_text()
                #   id
                dict_[col_id] = content_xml_bn.find(el_fin_ins_id).find(el_othr_id).find(
                    el_id).get_text()
                #   id - prtry
                dict_[col_id_prtry] = content_xml_bn.find(el_fin_ins_id).find(el_othr_id).find(
                    el_tp).find(el_pr_try).get_text()
                #   mercado
                dict_[col_mercado] = content_xml_bn.find(el_plc_of_listg).find(
                    el_mkt_id_rcd).get_text()
                #   quantidade de negócios
                try:
                    dict_[col_qtd_neg] = content_xml_bn.find(el_trad_dtls).find(
                        el_trad_qty).get_text()
                except:
                    dict_[col_qtd_neg] = qtd_erro
                #   string exercício, notional negociado brl
                for col_, key_ in [
                    (col_str_exercicio, el_mkt_data_strm_id),
                    (col_notional_neg_dom, el_nt_fin),
                    (col_notional_neg_int, el_int_fin),
                    (col_qtd_contratos_aberto, el_opn_instr),
                    (col_qtd_contratos_neg, el_fin_instrm_qt),
                    (col_melhor_ask, el_best_ask),
                    (col_prim_prec, el_first_pric),
                    (col_menor_preco, el_min_pric),
                    (col_maior_prec, el_max_prec),
                    (col_preco_med, el_avg_pric),
                    (col_ult_preco, el_last_pric),
                    (col_qtd_negocios_sess_reg, el_rglr_qt),
                    (col_qtd_negociada_sess_reg, el_rglr_trad),
                    (col_notional_neg_dom_sess_reg, el_ntl_rglr_vol),
                    (col_notional_neg_int_sess_reg, el_intl_rglr_vol),
                    (col_pct_osc_int, el_oscn_pctg)
                ]:
                    try:
                        dict_[col_] = content_xml_bn.find(el_fin_att).find(key_).get_text()
                    except:
                        dict_[col_] = qtd_erro
                #   adicionando dicionário a lista serializada
                list_ser.append(dict_)
        # adicionando a dataframe de exportação
        df_trd_rpt = pd.DataFrame(list_ser)
        # removendo eventuais duplicatas
        df_trd_rpt.drop_duplicates(inplace=True)
        # alterando tipo de colunas de interesse
        df_trd_rpt = df_trd_rpt.astype({
            col_dt_transacao: str,
            col_ticker: str,
            col_id: str,
            col_id_prtry: str,
            col_mercado: str,
            col_qtd_neg: int,
            col_str_exercicio: str,
            col_notional_neg_dom: float,
            col_notional_neg_int: float,
            col_qtd_contratos_aberto: int,
            col_qtd_contratos_neg: int,
            col_melhor_ask: float,
            col_prim_prec: float,
            col_menor_preco: float,
            col_maior_prec: float,
            col_preco_med: float,
            col_ult_preco: float,
            col_qtd_negocios_sess_reg: int,
            col_qtd_negociada_sess_reg: int,
            col_notional_neg_dom_sess_reg: float,
            col_notional_neg_int_sess_reg: float,
            col_pct_osc_int: float
        })
        # alterando tipo de colunas de data
        # [DatesBR().date_to_datetime(
        #         d, str_time='%H%M%S', format_dt='%Y-%m-%dT%H:%M:%Sz') for d in df_trd_rpt[col_].tolist()]
        for col_ in [col_dt_transacao]:
            df_trd_rpt[col_] = [DatesBR().str_date_to_datetime(d, formato_dt) for d in df_trd_rpt[
                col_].tolist()]
        # coletando colunas de interesse
        list_cols = list(df_trd_rpt.columns)
        # coletando tipo de mercado por ativo
        df_trd_rpt = df_trd_rpt.merge(
            df_, how='left', 
            left_on=[col_ticker], 
            right_on=[col_ticker_cd], 
            suffixes=('_', '')
        )
        list_cols.append(col_tipo_mercado_cd)
        # mantendo apenas colunas de interesse
        df_trd_rpt = df_trd_rpt[list_cols]
        # ordenando por ticker
        df_trd_rpt.sort_values([col_ticker], ascending=[False], inplace=True)
        # adding logging
        df_trd_rpt = DBLogs().audit_log(
            df_trd_rpt, 
            url,
            DatesBR().utc_from_dt(DatesBR().curr_date)
        )
        # retornando dataframe de interesse
        return df_trd_rpt

# print(self.cenarios_tipo_curva(list_cenarios_tipo_curva=[2939, 2940, 2941]))
# print(self.cenarios_risco_tipo_spot(list_cenarios_tipo_spot=[3012, 3134, 3135]))
# df_fpr_b3 = self.fprs_cenarios_curva_spot_merge
# print(df_fpr_b3)
# df_mtm_b3 = self.mtm_compra_venda()
# df_mtm_b3.to_excel(r"D:\Downloads\mtm-b3_20220928_1426.xlsx", index=False)
# print(self.collateral_acc_spot_bov_b3)
# df_fpr_b3 = self.fprs_cenarios_curva_spot_merge()
# df_fpr_b3.to_excel(
#     r"D:\Downloads\teste-fpr-b3_20220929_1832.xlsx", index=False)
# print(self.mtm_compra_venda())
# print(self.fatores_primitivos_risco_b3())



### VALIDATE PROPERTIES AND METHODS FROM A CLASS ###

from pydantic import validate_arguments


class ValidateAllMethodsMeta(type):
    '''
    DOCSTRING: VALIDATE ALL INPUT AND OUTPUT TYPES OF A CLASS
    INPUTS: NAME, BASES, DCT
    OUTPUTS: VALIDATED CLASS
    '''
    
    def __new__(cls, name, bases, dct):
        for attr_name, attr_value in dct.items():
            if callable(attr_value) and not attr_name.startswith("__"):
                dct[attr_name] = validate_arguments(attr_value)
        return super().__new__(cls, name, bases, dct)


if __name__ == '__main__':

    class MyClass(metaclass=ValidateAllMethodsMeta):

        def method_one(self, x: int, y: int) -> int:
            return x + y

        def method_two(self, z: str) -> str:
            return z.upper()


    # usage
    obj = MyClass()
    print(obj.method_one(3, 4))  # Outputs: 7
    try:
        print(obj.method_one("a", "b"))  # Raises ValidationError
    except Exception as e:
        print("Validation Error:", e)


### LIQUIDITTY RISK LIB

class LiquidityRatios:

    def current_ratio(self, float_curr_assets, float_curr_liabilities):
        '''
        REFERENCES: https://www.investopedia.com/terms/l/liquidityratios.asp
        DOCSTRING: MEASURES A COMPANY'S ABILITY TO PAY OFF ITS CURRENT LIABILITIES 
            (PAYABLE WITHIN ONE YEAR) WITH ITS TOTAL CURRENT ASSETS SUCH AS CASH, 
            ACCOUNTS RECEIVABLE, AND INVENTORIES
        INPUTS: CURRENT ASSETS AND LIABILITIES
        OUTPUTS: FLOAT
        '''
        return float_curr_assets / float_curr_liabilities

    def quick_ratio(self, float_curr_assets, float_inventories, float_curr_liabilities):
        '''
        REFERENCES: https://www.investopedia.com/terms/l/liquidityratios.asp
        DOCSTRING: MEASURES A COMPANY'S ABILITY TO MEET ITS SHORT-TERM OBLIGATIONS WITH ITS 
            MOST LIQUID ASSETS AND THEREFORE EXCLUDES INVENTORIES FROM ITS CURRENT ASSETS. 
            IT IS ALSO KNOWN AS THE ACID-TEST RATIO
        INPUTS:
        OUTPUTS: FLOAT
        '''
        return (float_curr_assets - float_inventories) / float_curr_liabilities
    
    def dso(self, float_avg_accounts_rec, float_rev, days=365):
        '''
        REFERENCES: https://www.investopedia.com/terms/l/liquidityratios.asp
        DOCSTRING: DAYS SALES OUTSTANDING (DSO) REFERS TO THE AVERAGE NUMBER OF DAYS IT TAKES A 
            COMPANY TO COLLECT PAYMENT AFTER IT MAKES A SALE. A HIGH DSO MEANS THAT A COMPANY 
            IS TAKING UNDULY LONG TO COLLECT PAYMENT AND IS TYING UP CAPITAL IN RECEIVABLES. 
            DSOS ARE GENERALLY CALCULATED ON A QUARTERLY OR ANNUAL BASIS
        INPUTS:
        OUTPUTS: FLOAT
        '''
        return float(days) / (float_rev / float_avg_accounts_rec)

    def cash_ratio(self, float_cash_eq, float_curr_liabilities):
        '''
        REFERENCES: https://www.toppr.com/guides/principles-and-practice-of-accounting/accounting-ratios/liquidity-ratios/
        DOCSTRING: MEASURES THE ABSOLUTE LIQUIDITY OF THE FIRM. IT MEASURES WHETHER A FIRM 
            CAN PAY THE CURRENT DEBTS BY USING ONLY THE CASH BALANCES, BANK BALANCES AND 
            MARKETABLE SECURITIES
        INPUTS:
        OUTPUTS: FLOAT
        '''
        return float_cash_eq / float_curr_liabilities


class SolvencyRatios:

    def interest_coverage_ratio(self, float_ebit, float_int_exp):
        '''
        REFERENCES: https://www.investopedia.com/terms/s/solvencyratio.asp,
        DOCSTRING: THE INTEREST COVERAGE RATIO MEASURES HOW MANY TIMES A COMPANY CAN COVER 
            ITS CURRENT INTEREST PAYMENTS WITH ITS AVAILABLE EARNINGS. IN OTHER WORDS, 
            IT MEASURES THE MARGIN OF SAFETY A COMPANY HAS FOR PAYING INTEREST ON ITS DEBT 
            DURING A GIVEN PERIOD. THE HIGHER THE RATIO, THE BETTER. IF THE RATIO FALLS TO 
            1.5 OR BELOW, IT MAY INDICATE THAT A COMPANY WILL HAVE DIFFICULTY MEETING 
            THE INTEREST ON ITS DEBTS.
        INPUTS:
        OUTPUTS: FLOAT 
        '''
        return float_ebit / float_int_exp
    
    def debt_to_assets_ratio(self, float_debt, float_assets):
        '''
        REFERENCES: https://www.investopedia.com/terms/s/solvencyratio.asp
        DOCSTRING: THE DEBT-TO-ASSETS RATIO MEASURES A COMPANY'S TOTAL DEBT TO ITS TOTAL ASSETS. 
            IT MEASURES A COMPANY'S LEVERAGE AND INDICATES HOW MUCH OF THE COMPANY IS FUNDED 
            BY DEBT VERSUS ASSETS, AND THEREFORE, ITS ABILITY TO PAY OFF ITS DEBT WITH 
            ITS AVAILABLE ASSETS. A HIGHER RATIO, ESPECIALLY ABOVE 1.0, INDICATES THAT A COMPANY 
            IS SIGNIFICANTLY FUNDED BY DEBT AND MAY HAVE DIFFICULTY MEETINGS ITS OBLIGATIONS.
        INPUTS:
        OUTPUTS:
        '''
        return float_debt / float_assets
    
    def equiy_ratio(self, float_tse, float_assets):
        '''
        REFERENCES: https://www.investopedia.com/terms/s/solvencyratio.asp
        DOCSTRING: THE EQUITY RATIO, OR EQUITY-TO-ASSETS, SHOWS HOW MUCH OF A COMPANY IS FUNDED 
            BY EQUITY AS OPPOSED TO DEBT. THE HIGHER THE NUMBER, THE HEALTHIER A COMPANY IS. 
            THE LOWER THE NUMBER, THE MORE DEBT A COMPANY HAS ON ITS BOOKS RELATIVE TO EQUITY.
        INPUTS: TOTAL SHAREHOLDER'S EQUITY AND TOTAL ASSETS
        OUTPUTS: FLOAT
        '''
        return float_tse / float_assets
    
    def debt_to_equity_ratio(self, float_debt, float_equity):
        '''
        REFERENCES: https://www.investopedia.com/terms/s/solvencyratio.asp
        DOCSTRING: THE D/E RATIO IS SIMILAR TO THE DEBT-TO-ASSETS RATIO, IN THAT IT INDICATES HOW 
            A COMPANY IS FUNDED, IN THIS CASE, BY DEBT. THE HIGHER THE RATIO, THE MORE DEBT 
            A COMPANY HAS ON ITS BOOKS, MEANING THE LIKELIHOOD OF DEFAULT IS HIGHER. THE RATIO 
            LOOKS AT HOW MUCH OF THE DEBT CAN BE COVERED BY EQUITY IF THE COMPANY NEEDED TO LIQUIDATE.
        INPUTS:
        OUTPUTS:
        '''
        return float_debt / float_equity

    def altmans_z_score(self, float_nwk_cap, float_total_assets, float_ret_earnings, float_ebit, 
                        float_mkt_cap, float_total_liabilities, float_sales):
        '''
        REFERENCES: https://medium.com/quant-factory/calculating-altman-z-score-with-python-3c6697ee7aee
        DOCSTRING: THE ALTMAN Z-SCORE QUANTIFIES THE FINANCIAL STABILITY OF A COMPANY TO PREDICT 
            HOW LIKELY A COMPANY WILL BECOME INSOLVENT - A SCORE BELOW 1.8 MEANS IT’S LIKELY THE 
            COMPANY IS HEADED FOR BANKRUPTCY, WHILE COMPANIES WITH SCORES ABOVE 3 ARE NOT LIKELY 
            TO GO BANKRUPT
        INPUTS: WORKING CAPITAL, TOTAL ASSETS, RETAINED EARNINGS, EBIT, MARKET CAPITALISATION, 
            TOTAL LIABILITIES AND SALES 
        OUTPUTS:
        '''
        return 1.2 * (float_nwk_cap / float_total_assets) \
            + 1.4 * (float_ret_earnings / float_total_assets) \
            + 3.3 * (float_ebit / float_total_assets) \
            + 0.6 * (float_mkt_cap / float_total_liabilities) \
            + 1.0 * (float_sales / float_total_assets)


### HANDLING OBJECTS OF A MYRIAD OF DATATYPES ###

import ast
from stpstone.handling_data.str import StrHandler


class HandlingObjects:

    def literal_eval_data(self, data_object, str_left_bound=None, str_right_bound=None):
        '''
        DOCSTRING: LITERAL EVAL AN OBJECT TO ITS INHERENT TYPE, FOR INSTANCE A STRING IN LIST
            OR DICT FORMAT
        INPUTS: DATA OBJECT, STRING LEFT BOUND, STRING RIGHT BOUND
        OUTPUTS:
        '''
        if any([x == None for x in [str_left_bound, str_right_bound]]):
            return ast.literal_eval(data_object)
        else:
            return ast.literal_eval(StrHandler().find_between(
                str(data_object), str_left_bound, str_right_bound))



### COLLECT WORLD GOVERNMENT BONDS DATA

import pandas as pd
from datetime import datetime, timezone
from stpstone.settings._global_slots import YAML_WGBD
from stpstone.handling_data.html import HtmlHndler
from stpstone.handling_data.dicts import HandlingDicts
from stpstone.handling_data.str import StrHandler
from stpstone.loggs.db_logs import DBLogs
from stpstone.cals.handling_dates import DatesBR


class WGBD:

    def parse_utc_dt_str(self, str_dt):
        '''
        DOCSTRING: PARSER UTC DATETIME TO UTC FORMAT
        INTPUTS: STRING DATETIME
        OUTPUTS: DATETIME
        '''
        # parse the date and time string without the timezone info
        dt_obj = datetime.strptime(str_dt[:-6], '%d %b %Y %H:%M')
        # manually set the UTC timezone
        dt_utc = dt_obj.replace(tzinfo=timezone.utc)
        # return datetime
        return dt_utc

    @property
    def bonds_10y_rates_rating(self):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # setting variables
        list_ser = list()
        # request html
        bs_html = HtmlHndler().html_bs_parser(
            YAML_WGBD['10y_rr']['url'], 
            bl_verify=YAML_WGBD['10y_rr']['bl_verify']
        )
        # getting last update date
        str_last_update = StrHandler().get_string_after_substr(
            bs_html.find('article').find(
                'div', class_=YAML_WGBD['10y_rr']['class_div_last_update']
            ).get_text(), 
            YAML_WGBD['10y_rr']['data_last_update']
        ).strip()
        # looping within tables
        for bs_table in bs_html.find_all('table'):
            #   collecting raw headers
            list_th_1 = [
                el.get_text()\
                    .replace('\n ', '')\
                    .replace('\n', '')\
                    .strip()\
                    .replace(' ', '_')\
                    .replace('-', '')\
                    .replace('▴', '')\
                    .replace('&', '_AND_')\
                    .upper()
                for el in bs_table.find('thead').find_all('tr')[0].find_all('div')
            ]
            list_th_2 = [
                el.get_text()\
                    .replace('\n ', '')\
                    .replace('\n', '')\
                    .strip()\
                    .replace(' ', '_')\
                    .replace('-', '')\
                    .replace('▴', '')\
                    .replace('&', '_AND_')\
                    .upper()
                for el in bs_table.find('thead').find_all('tr')[1].find_all('th')
            ]
            #   building consolidated headers' list
            list_th = [
                list_th_2[1], 
                list_th_1[0] + '_' + list_th_2[2],
                list_th_1[1] + '_' + list_th_2[3],
                list_th_1[2] + '_' + list_th_2[5],
                list_th_1[3] + '_' + list_th_2[6],
                list_th_1[3] + '_' + list_th_2[7],
                list_th_1[3] + '_' + list_th_2[8]
            ]
            #   looping within rows in tbody
            for bs_tr in bs_table.find('tbody').find_all('tr'):
                #   get all td data
                list_td = [
                    el.get_text()\
                        .replace('\n', '')\
                        .replace(' (*)', '')\
                        .replace('\t', '')\
                        .strip()
                    for el in bs_tr.find_all('td')
                    if (len(el.get_text()) > 1) & (el.get_text() != '')
                ]
                list_td = [x for x in list_td if len(x) > 0]
                #   case when len of tds is less than ths, then check the s&p rating, if is not a 
                #       string add an error tag
                if len(list_th) > len(list_td):
                    if StrHandler().bl_has_numbers(list_td[1]) == True:
                        list_td.insert(1, YAML_WGBD['10y_rr']['data_error'])
                    else:
                        raise Exception('Error - data missing in tbody unable to replace, please '
                                        + 'validate the code')
                #   creating list of dictionaries
                list_ser = HandlingDicts().pair_headers_with_data(
                    list_th, 
                    list_td
                )
        # creating dataframe
        df_b10_rr = pd.DataFrame(list_ser)
        # remove duplicates
        df_b10_rr.drop_duplicates(inplace=True)
        # adding logging
        df_b10_rr = DBLogs().audit_log(
            df_b10_rr, YAML_WGBD['10y_rr']['url'], 
            self.parse_utc_dt_str(str_last_update)
        )
        # returning dataframe
        return df_b10_rr
    
    @property
    def inv_yc_economies(self):
        '''
        DOCSTRING: GET INVERTED YIELD CURVE ECONOMIES
        INPUTS: -
        OUTPUTS: TUPLE OF DATAFRAMES
        '''
        # request html
        bs_html = HtmlHndler().html_bs_parser(
            YAML_WGBD['inv_yc_ec']['url'], 
            bl_verify=YAML_WGBD['inv_yc_ec']['bl_verify']
        )
        # getting last update date
        str_last_update = StrHandler().get_string_after_substr(
                bs_html.find('article').find(
                    'p', class_='-f14').get_text(), 
                YAML_WGBD['inv_yc_ec']['data_last_update']
            ).strip()
        # looping within tables
        for i_tb, bs_table in enumerate(bs_html.find_all('table')):
            #   dealing with first table
            if i_tb == 0:
                #   reseting variables
                list_ser = list()
                #   looping within rows in tbody
                for bs_tr in bs_table.find('tbody').find_all('tr'):
                    #   collecting countries, rate and level of invertion
                    list_ = [
                        el.get_text()\
                            .replace('\n', '')\
                            .strip()
                        for el in bs_tr.find_all('td')
                    ]
                    #   catching current rate
                    str_rating = list_[0]
                    #   consolidating level of invertion
                    for str_lvl_inv, list_ in [
                        (YAML_WGBD['inv_yc_ec']['data_tot_inv'], 
                         [x.strip() for x in list_[1].split(YAML_WGBD['inv_yc_ec'][
                             'split_countries_lvl_inv'])]), 
                        (YAML_WGBD['inv_yc_ec']['data_prt_inv'], 
                         [x.strip() for x in list_[1].split(YAML_WGBD['inv_yc_ec'][
                             'split_countries_lvl_inv'])]), 
                        (YAML_WGBD['inv_yc_ec']['data_min_inv'], 
                         [x.strip() for x in list_[1].split(YAML_WGBD['inv_yc_ec'][
                             'split_countries_lvl_inv'])]), 
                    ]:
                        list_ser.extend([
                            {
                                YAML_WGBD['inv_yc_ec']['col_rating']: str_rating,
                                YAML_WGBD['inv_yc_ec']['col_country']: x,
                                YAML_WGBD['inv_yc_ec']['col_lvl_inv']: str_lvl_inv
                            } for x in list_
                        ])
                #   import list serialized to dataframe
                df_inv_yc_ec = pd.DataFrame(list_ser)
                # remove duplicates
                df_inv_yc_ec.drop_duplicates(inplace=True)
            #   dealing with first table
            elif i_tb == 1:
                #   reseting variables
                list_ser = list()
                #   collecting countries and yield spreads headers
                list_th = [
                    el.get_text()\
                        .replace('\n', '')\
                        .replace('Long vs Short Term', '')\
                        .replace('Mid vs Short Term', '')\
                        .replace('Short Term', '')\
                        .replace(' ', '_')\
                        .strip()\
                        .upper()
                    for el in bs_table.find('thead').find_all('th')
                    if len(el.get_text()) > 0
                ]
                #   collecting countries and yield spreads data
                list_td = [
                    #   collecting bps and transforming to float
                    float(
                        el.get_text()\
                            .replace('\n', '')\
                            .replace(YAML_WGBD['inv_yc_ec']['data_bp'], '')\
                            .strip()
                    ) * 1E-4
                    if StrHandler().match_string_like(
                        el.get_text()\
                            .replace('\n', '')\
                            .strip(), 
                        '* bp'
                    ) == True
                    #   collecting other data, such as country name
                    else el.get_text()\
                            .replace('\n', '')\
                            .strip()
                    for el in bs_table.find('tbody').find_all('td')
                    if len(el.get_text()\
                            .replace('\n', '')\
                            .strip()
                    ) > 0
                ]
                #   creating list of dictionaries
                list_ser = HandlingDicts().pair_headers_with_data(
                    list_th, 
                    list_td
                )
                #   import list serialized to dadtaframe
                df_inv_yc_spreads = pd.DataFrame(list_ser)
                #   remove duplicates
                df_inv_yc_spreads.drop_duplicates(inplace=True)
            else:
                raise Exception('Table not identified, of inverted yield curve countries, please '
                                + 'revisit the code')
        #   adding logging columns
        for df_ in [
            df_inv_yc_ec,
            df_inv_yc_spreads
        ]:
            #   adding logging
            df_ = DBLogs().audit_log(
                df_, YAML_WGBD['inv_yc_ec']['url'], 
                self.parse_utc_dt_str(str_last_update)
            )
        #   returning dataframes
        return df_inv_yc_ec, df_inv_yc_spreads
    
    @property
    def yield_spreads_nations(self):
        '''
        DOCSTRING: GET YIELD SPREADS AGAINST GERMANY, USA, CHINA AND AUSTRALIA
        INPUTS: -
        OUTPUTS: DATAFRAME
        '''
        # setting variables
        list_td = list()
        # request html
        bs_html = HtmlHndler().html_bs_parser(
            YAML_WGBD['yield_spreads_nations']['url'], 
            bl_verify=YAML_WGBD['yield_spreads_nations']['bl_verify']
        )
        # getting last update date
        str_last_update = StrHandler().get_string_after_substr(
            [
                el.get_text() for el in bs_html.find('article').find_all('p')
                if StrHandler().match_string_like(
                    el.get_text(),
                    YAML_WGBD['yield_spreads_nations']['data_last_update_like']
                )
            ][0], 
            YAML_WGBD['yield_spreads_nations']['data_last_update']
        ).strip()
        # getting headers
        list_th1 = [
            el.get_text()\
                .replace(' ', '_')\
                .strip()\
                .upper()
            for el in bs_html.find('table').find('thead').find_all('tr')[0].find_all('th')
            if len(el.get_text()) > 0
        ]
        list_th2 = [
            el.get_text()\
                .replace('▴', '')\
                .replace('\n', '')\
                .strip()\
                .upper()\
                .replace(' ', '_')
            for el in bs_html.find('table').find('thead').find_all('tr')[1].find_all('th')
            if len(el.get_text()) > 0
        ]
        list_th = [list_th2[0], list_th2[1]]
        list_th.extend([
            list_th1[0] + '_' + x for x in list_th2[2:]
        ])
        # getting data
        for el in bs_html.find('table').find('tbody').find_all('td'):
            #   ignoring missing data
            if len(
                el.get_text()\
                    .replace('\n', '')\
                    .strip()
            ) == 0: continue
            #   collecting bps data
            elif StrHandler().match_string_like(
                    el.get_text()\
                        .replace('\n', '')\
                        .strip(), 
                    '* bp'
                ) == True:
                list_td.append(float(
                    el.get_text()\
                        .replace('\n', '')\
                        .replace(YAML_WGBD['yield_spreads_nations']['data_bp'], '')\
                        .strip()
                ) * 1E-4)
            #   collecting 10y data
            elif StrHandler().match_string_like(
                    el.get_text()\
                        .replace('\n', '')\
                        .strip(), 
                    '*%'
            ) == True:
                list_td.append(float(
                    el.get_text()\
                        .replace('\n', '')\
                        .replace(YAML_WGBD['yield_spreads_nations']['data_pct'], '')\
                        .strip()
                ) * 1E-2)
            #   collecting other data, such as country name
            else:
                list_td.append(
                    el.get_text()\
                        .replace('\n', '')\
                        .strip()
                )
        # creating list of dictionaries
        list_ser = HandlingDicts().pair_headers_with_data(
            list_th, 
            list_td
        )
        # import list serialized to dadtaframe
        df_spreads_nations = pd.DataFrame(list_ser)
        # remove duplicates
        df_spreads_nations.drop_duplicates(inplace=True)
        # adding logging
        df_spreads_nations = DBLogs().audit_log(
            df_spreads_nations, YAML_WGBD['yield_spreads_nations']['url'], 
            self.parse_utc_dt_str(str_last_update)
        )
        # returning dataframe
        return df_spreads_nations
    
    @property
    def ratings(self):
        '''
        DOCSTRING: GET RATINGS FROM THE MAIN AGENCIES
        INPUTS: -
        OUTPUTS: DATAFRAME
        '''
        # setting variables
        list_td = list()
        # request html
        bs_html = HtmlHndler().html_bs_parser(
            YAML_WGBD['yield_spreads_nations']['url'], 
            bl_verify=YAML_WGBD['yield_spreads_nations']['bl_verify']
        )
        # getting last update date
        str_last_update = StrHandler().get_string_after_substr(
            [
                el.get_text() for el in bs_html.find('article').find_all('p')
                if StrHandler().match_string_like(
                    el.get_text(),
                    YAML_WGBD['yield_spreads_nations']['data_last_update_like']
                )
            ][0], 
            YAML_WGBD['yield_spreads_nations']['data_last_update']
        ).strip()
        # request html
        bs_html = HtmlHndler().html_bs_parser(
            YAML_WGBD['credit_rating_nations']['url'], 
            bl_verify=YAML_WGBD['credit_rating_nations']['bl_verify']
        )
        # getting headers
        list_th1 = [
            el.get_text()\
                .replace(' ', '_')\
                .replace('▴', '')\
                .replace('&', '_AND_')\
                .replace("'", '')\
                .strip()\
                .upper()
            for el in bs_html.find('table').find('thead').find_all('th')
            if len(el.get_text()) > 0
        ]
        list_th = [list_th1[0]]
        for i in range(1, len(list_th1)):
            list_th.append(list_th1[i])
            list_th.append(list_th1[i] + '_OUTLOOK')
        # getting data
        for el in bs_html.find('table').find('tbody').find_all('td'):
            #   check whether info is empty, in this case continue
            if len(el.get_text().strip()) == 0: 
                bl_country = True
                continue
            #   appending info regarding country and current rating
            list_td.append(el.get_text()\
                .strip())
            #   getting info regarding outlook
            try:
                list_td.append(
                    el.find('i')['title']\
                        .replace(' ', '_')\
                        .upper()
                )
            except (AttributeError, TypeError):
                if bl_country == False:
                    list_td.append(
                        'NEUTRAL_OUTLOOK'
                    )
                else:
                    bl_country = False
        # creating list of dictionaries
        list_ser = HandlingDicts().pair_headers_with_data(
            list_th, 
            list_td
        )
        # import list serialized to dadtaframe
        df_ratings = pd.DataFrame(list_ser)
        # remove duplicates
        df_ratings.drop_duplicates(inplace=True)
        # adding logging
        df_ratings = DBLogs().audit_log(
            df_ratings, YAML_WGBD['credit_rating_nations']['url'], 
            self.parse_utc_dt_str(str_last_update)
        )
        # returning dataframe
        return df_ratings


### ANBIMA FILES ###

import pandas as pd
from requests import request
from stpstone.settings._global_slots import YAML_ANBIMA
from stpstone.loggs.create_logs import CreateLog
from stpstone.handling_data.lists import HandlingLists
from stpstone.cals.handling_dates import DatesBR
from stpstone.handling_data.json import JsonFiles
from stpstone.handling_data.folders import DirFilesManagement
from stpstone.handling_data.txt import HandlingTXTFiles
from stpstone.handling_data.str import StrHandler
from stpstone.finance.b3.up2data_web import UP2DATAB3
from stpstone.handling_data.html import HtmlHndler
from stpstone.handling_data.xml import XMLFiles
from stpstone.finance.derivatives.options.european import EuropeanOptions
from stpstone.finance.b3.market_data import MDB3
from stpstone.loggs.db_logs import DBLogs


class AnbimaMTM:

    def general_req(self, url, bl_verify, method='GET', str_sep='@'):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # setting variables
        list_ser = list()
        # requesting response
        resp_req = request(method, url, bl_verify)
        # raise exception in case of status code different from 2xx
        resp_req.raise_for_status()
        # turning content into list of rows
        str_tb = resp_req.text
        list_tb = str_tb.split('\r\n')[2:]
        # turn into serialized list
        for row in list_tb:
            #   reseting variables
            dict_ = dict()
            #   defining header
            if row.split(str_sep)[0] == 'Titulo':
                list_headers = row.split(str_sep)
            #   appending data
            else:
                #   creating serialized list
                dict_ = dict(zip(list_headers, row.replace(',', '.').split(str_sep)))
                #   appending to consolidated list
                list_ser.append(dict_)
        # turning into dataframe
        df_ = pd.DataFrame(list_ser)
        # droping nan values
        df_.dropna(inplace=True)
        # returning dataframe
        return df_

    def br_treasury_bonds(self, wd_bef=1):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # url
        url = YAML_ANBIMA['sec_mkt_prcs']['br_treasuries']['url'].format(
            DatesBR().sub_working_days(DatesBR().curr_date, wd_bef).strftime(
                '%y%m%d'
            )
        )
        # requesting secondary market prices
        df_br_tb = self.general_req(url, YAML_ANBIMA['sec_mkt_prcs']['br_treasuries']['bl_verify'])
        # columns
        list_headers = list(df_br_tb.columns)
        # chaning data types
        df_br_tb = df_br_tb.astype({
            list_headers[0]: str,
            list_headers[1]: int,
            list_headers[2]: int,
            list_headers[3]: int,
            list_headers[4]: int,
            list_headers[5]: float,
            list_headers[6]: float,
            list_headers[7]: float,
            list_headers[8]: float,
            list_headers[9]: float,
            list_headers[10]: float,
            list_headers[11]: float,
            list_headers[12]: float,
            list_headers[13]: float,
            list_headers[14]: str
        })
        # adding logging
        df_br_tb = DBLogs().audit_log(
            df_br_tb, 
            url,
            DatesBR().utc_from_dt(
                DatesBR().sub_working_days(DatesBR().curr_date, wd_bef)
            )
        )
        # returning dataframe
        return df_br_tb
    
    def corporate_bonds(self, wd_bef=1):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # url
        url = YAML_ANBIMA['sec_mkt_prcs']['corporate_bonds']['url'].format(
            DatesBR().sub_working_days(DatesBR().curr_date, wd_bef).strftime(
                '%y%m%d'
            )
        )
        # requesting secondary market prices
        df_br_tb = self.general_req(url, YAML_ANBIMA['sec_mkt_prcs']['corporate_bonds']['bl_verify'])
        # columns
        list_headers = list(df_br_tb.columns)
        # chaning data types
        df_br_tb = df_br_tb.astype({
            list_headers[0]: str, ## code
            list_headers[1]: str, ## name
            list_headers[2]: str, ## date
            list_headers[3]: str, ## maturity
            list_headers[4]: str, ## index
            list_headers[5]: float, ## bid rate
            list_headers[6]: float, ## ask rate
            list_headers[7]: float, ## market reference rate
            list_headers[8]: float, ## standard deviation
            list_headers[9]: float, ## indicative min range
            list_headers[10]: float, ## indicative max range
            list_headers[11]: float, ## present value (pv)
            list_headers[12]: float, ## pv @ par / VNE (future value - fv - or valor nominal 
                                     ##     de emissão, pt-br)
            list_headers[13]: float, ## duration
        })
        # adding logging
        df_br_tb = DBLogs().audit_log(
            df_br_tb, 
            url,
            DatesBR().utc_from_dt(
                DatesBR().sub_working_days(DatesBR().curr_date, wd_bef)
            )
        )
        # returning dataframe
        return df_br_tb
    
    @property
    def ima(self, str_sep='@'):
        '''
        DOCSTRING: MARKET INDEXES ANBIMA
        INPUTS:
        OUTPUTS:
        '''
        # setting variables
        dict_list_ser = dict()
        # requesting response
        resp_req = request(
            YAML_ANBIMA['sec_mkt_prcs']['ima']['method'], 
            YAML_ANBIMA['sec_mkt_prcs']['ima']['url'], 
            YAML_ANBIMA['sec_mkt_prcs']['ima']['bl_verify']
        )
        # raise exception in case of status code different from 2xx
        resp_req.raise_for_status()
        # turning content into list of rows
        str_tb = resp_req.text
        list_tb = str_tb.split('\r\n')[2:]
        # turn into serialized list
        for row in list_tb:
            #   table identifier
            str_tb_id = str(row.split(str_sep)[0])
            #   reseting variables
            dict_ = dict()
            #   if index dataframe not in dict of serialized lists, create a key
            if str_tb_id not in dict_list_ser:
                dict_list_ser[str_tb_id] = list()
            #   defining header
            if row.split(str_sep)[1] == 'Data de Referência':
                list_headers = row.split(str_sep)
            #   appending data
            else:
                #   creating serialized list
                dict_ = dict(zip(list_headers, row.replace(',', '.').split(str_sep)))
                #   appending to consolidated list
                dict_list_ser[str_tb_id].append(dict_)
        # turning into dataframes
        df_ima_pvs = pd.DataFrame(dict_list_ser['1'])
        df_ima_th_portf = pd.DataFrame(dict_list_ser['2'])
        # adding logging
        for df_ in [df_ima_pvs, df_ima_th_portf]:
            df_ = DBLogs().audit_log(
                df_, 
                YAML_ANBIMA['sec_mkt_prcs']['ima']['url'],
                DatesBR().utc_from_dt(DatesBR().curr_date)
            )
        # returning dataframes
        return df_ima_pvs, df_ima_th_portf


### ANBIMA STATISTICS ###

import pandas as pd
from math import nan
from pprint import pprint
from stpstone.settings._global_slots import YAML_ANBIMA
from stpstone.cals.handling_dates import DatesBR
from stpstone.handling_data.json import JsonFiles
from stpstone.handling_data.str import StrHandler
from stpstone.finance.comdinheiro.api_request import ComDinheiro
from stpstone.handling_data.object import HandlingObjects
from stpstone.handling_data.lists import HandlingLists
from stpstone.document_numbers.br import DocumentsNumbersBR
from stpstone.handling_data.dicts import HandlingDicts
from stpstone.handling_data.html import HtmlHndler
from stpstone.loggs.db_logs import DBLogs


class AnbimaStats:

    @property
    def pmi_hat(self):
        '''
        DOCSTRING: IPCA PMI PROJECTED BY ANBIMA
        INPUTS: -
        OUTPUTS: DATAFRAME
        '''
        # setting variables
        list_ser = list()
        # request html
        bs_html = HtmlHndler().html_bs_parser(
            YAML_ANBIMA['anbima_stats']['url_price_indexes'], 
            bl_verify=YAML_ANBIMA['anbima_stats']['bl_verify']
        )
        # looping within tables
        for i_tb, bs_table in enumerate(bs_html.find_all('table')):
            #   reseting variables
            list_data = list()
            list_headers = list()
            int_len_th = list()
            int_ref_tb = i_tb % 3
            int_ref_idx = i_tb//3
            #   defining current index
            if int_ref_idx == 0:
                str_idx = 'igpm'
            elif int_ref_idx == 1:
                str_idx = 'ipca'
            else:
                str_idx = 'others'
            #   defining the key of serialized list, accordingly to table name
            if int_ref_tb == 0:
                data_hat = YAML_ANBIMA['anbima_stats']['data_hat_mp1']
            elif int_ref_tb == 1:
                data_hat = YAML_ANBIMA['anbima_stats']['data_hat_mp2']
            elif int_ref_tb == 2:
                data_hat = YAML_ANBIMA['anbima_stats']['data_hat_l12m']
            #   looping within rows
            for i_tr, bs_tr in enumerate(bs_table.find_all('tr')):
                #   first row as header, others as 
                if i_tr == 1:
                    list_headers = [
                        StrHandler().remove_diacritics(x.get_text())\
                            .replace('*', '')\
                            .replace(' (%)', '')\
                            .replace(' da ', ' ')\
                            .replace(' de ', ' ')\
                            .replace(' ', '_')\
                            .replace('-', '')\
                            .upper()\
                            .replace(YAML_ANBIMA['anbima_stats']['col_ipca_eft'], 
                                     YAML_ANBIMA['anbima_stats']['col_pmi_eft'])\
                            .replace(YAML_ANBIMA['anbima_stats']['col_igpm_eft'], 
                                     YAML_ANBIMA['anbima_stats']['col_pmi_eft'])
                        for x in bs_tr.find_all('td')
                    ]
                    int_len_th = len(list_headers)
                    list_headers.extend([YAML_ANBIMA['anbima_stats']['col_pmi_core'], 
                                         YAML_ANBIMA['anbima_stats']['col_period_prediction']])
                elif i_tr > 0:
                    #   creating list to be appended in the consolidated data list
                    list_ = [
                        x.get_text().replace(',', '.')
                        for x in bs_tr.find_all('td')
                    ]
                    #   in case of length less than the headers, fill extremes with nan
                    if len(list_) < int_len_th:
                        list_.insert(0, nan)
                        list_.insert(len(list_headers)-1, nan)
                    #   adding pmi core
                    list_.extend([str_idx.upper(), data_hat.upper()])
                    #   extending to consolidated list
                    list_data.extend(list_)
            # creating serialized list
            list_ser = HandlingDicts().pair_headers_with_data(list_headers, list_data)
            # appending to exporting list
            list_ser.extend(list_ser)
        # importing to dataframe
        df_pmi_hat = pd.DataFrame(list_ser)
        # removing duplicates
        df_pmi_hat.drop_duplicates(inplace=True)
        # replacing values like nan
        df_pmi_hat.replace('-', nan, inplace=True)
        df_pmi_hat[YAML_ANBIMA['anbima_stats']['col_expirtation_dt']].fillna(
            YAML_ANBIMA['anbima_stats']['dt_fillna'], inplace=True)
        # adding request date
        df_pmi_hat[YAML_ANBIMA['anbima_stats']['col_request_date'].upper()] = \
            DatesBR().curr_date
        # altering data types
        for col_ in [
            YAML_ANBIMA['anbima_stats']['col_dt'], 
            YAML_ANBIMA['anbima_stats']['col_expirtation_dt']
        ]:
            df_pmi_hat[col_] = [DatesBR().str_date_to_datetime(x, 'DD/MM/YY') 
                                for x in df_pmi_hat[col_]]
        # sorting dataframe
        df_pmi_hat.sort_values([
            YAML_ANBIMA['anbima_stats']['col_pmi_core'], 
            YAML_ANBIMA['anbima_stats']['col_period_prediction'], 
            YAML_ANBIMA['anbima_stats']['col_dt']
        ], ascending=[False, True, False], inplace=True)
        # forward fill method (for nan values replace with last data, for the same column)
        df_pmi_hat = df_pmi_hat.groupby([
            YAML_ANBIMA['anbima_stats']['col_pmi_core'], 
            YAML_ANBIMA['anbima_stats']['col_period_prediction']
        ], as_index=False).apply(lambda group: group.ffill())
        df_pmi_hat.fillna(0, inplace=True)
        # altering data types
        df_pmi_hat = df_pmi_hat.astype({
            YAML_ANBIMA['anbima_stats']['col_month_collection']: str,
            YAML_ANBIMA['anbima_stats']['col_proj']: float,
            YAML_ANBIMA['anbima_stats']['col_pmi_core']: str,
            YAML_ANBIMA['anbima_stats']['col_period_prediction']: str,
            YAML_ANBIMA['anbima_stats']['col_month_proj']: str,
            YAML_ANBIMA['anbima_stats']['col_pmi_eft']: float
        })
        # creating logg
        df_pmi_hat = DBLogs().audit_log(
            df_pmi_hat, 
            YAML_ANBIMA['anbima_stats']['url_price_indexes'], 
            DatesBR().curr_date
        )
        # returning list
        return df_pmi_hat


### FINANCIAL MATH LIB ###

import numpy_financial as npf


class FinancialMath:

    def compound_interest(self, nominal_interest_rate, nominal_nper, real_nper):
        '''
        DOCSTRING: COMPOUND INTEREST (NOMINAL FEE)
        INPUTS: NOMINAL FEE, NPER REALTED TO NOMINAL FEE AND NPER REGARDING THE REAL FEE 
            FOR THE OBSERVED PERIOD OF TIME
        OUTPUTS: FLOAT
        '''
        # correcting parameters:
        if type(nominal_interest_rate) == list:
            nominal_interest_rate = 0.0
        # returning real compunded interest
        return float((1 + nominal_interest_rate) ** (real_nper / nominal_nper)) - 1.0

    def simple_interest_rate(self, nominal_interest_rate, nominal_nper, real_nper):
        '''
        DOCSTRING: SIMPLETE INTEREST RATE
        INPUTS: NOMINAL INTEREST RATE, NOMINAL NPER, REAL NPER
        OUTPUTS: FLOAT
        '''
        return 1 + nominal_interest_rate * nominal_nper / real_nper

    def present_value(self, rate, nper, pmt, fv=0, when='end'):
        '''
        DOCSTRING: PRESENT VALUE
        INPUTS: RATE (FLOAT), NPER (INT), PMT (PAYMENT, INT), FV (FUTURE VALUE, WHETHER INT OR 
            FLOAT), WHEN (BEGIN OR END OF PERIOD)
        OUTPUTS: FLOAT OF PV
        '''
        return npf.pv(rate, nper, pmt, fv, when)

    def future_value(self, rate, nper, pmt, pv=0, when='end'):
        '''
        DOCSTRING: FUTURE VALUE
        INPUTS: RATE (FLOAT), NPER (INT), PMT (PAYMENT, INT), PV (PRESENT VALUE, WHETHER INT OR 
            FLOAT), WHEN (BEGIN OR END OF PERIOD)        
        OUTPUTS: FLOAT OF PV
        '''
        return npf.fv(rate, nper, pmt, pv, when)

    def internal_rate_return(self, list_cash_flow):
        '''
        DOCSTRING: INTERNAL RATE OF RETURN
        INPUTS: LIST OF VALUES THAT COMPRISE THE CASH FLOW
        OUTPUTS: FLOAT OF IRR
        '''
        return npf.irr(list_cash_flow)

    def net_present_value(self, rate, list_cash_flow):
        '''
        DOCSTRING: NET PRESENT VALUE FOR A CASH FLOW DISCOUNTED BY A RATE; ALONG VALUES OUGHT BE 
            AT LEAST ONE NEGATIVE AND ONE POSITIVE
        INPUTS: RATE (FLOAT) AND VALUES 
        '''
        return npf.npv(rate, list_cash_flow)


# print(MathFinance().present_value(0.1, 1, 0, 110))
# # output
# -99.99999999999999

# print(MathFinance().internal_rate_return([-133, 85.3, 12.47, 55.23, 11.47]))
# # output
# 0.11863351188349469

# print(MathFinance().future_value(0.1, 12, 0, -8000))
# # output
# 25107.427013768025

# print(MathFinance().future_value(0.1, 1, 0, -100))
# # output
# 110.00000000000001



### HANDLING LXML TO EXTRACT HTML DATA ###

from lxml import html
from requests import request


class HandlingLXML:

    def fetch(self, url, method='get'):
        '''
        REFENCES: https://stackoverflow.com/questions/26944078/extracting-value-of-url-source-by-xpath-in-python
        DOCSTRING: FETCHING HTML DOCUMENT TO BE SELECTED BY XPATH
        INPUTS: URL
        OUTPUTS: -
        '''
        content = request(method, url).content
        tree = html.fromstring(content)
        return tree



### HANDLING LISTS ###

import re
import bisect
import numpy as np
from heapq import heappush, heappop
from collections import OrderedDict, Counter
from itertools import chain, tee, product, combinations
from joblib import Parallel, delayed
from stpstone.handling_data.json import JsonFiles
from stpstone.handling_data.str import StrHandler
from stpstone.handling_data.numbers import NumHandler


class HandlingLists:

    def get_first_occurrence_within_list(self, list_, obj_occurrence=None, bl_uppercase=False,
                                         bl_last_uppercase_before_capitalized=False, int_error=-1,
                                         int_error_obj_occurrence=-2, bl_regex_alphanumeric_false=False,
                                         bl_ignore_sole_letter=True,
                                         str_original_replace_1=',',
                                         str_original_replace_2='.', str_result_replace='',
                                         bl_audit=False):
        '''
        DOCSTRING: GET THE FIRST OCCURRENCE OF AN OBJECT WITHIN A LIST
        INPUTS: LIST, OBJECT AND BOOLEAN UPPERCASE (FALSE AS DEFAULT)
        OUTPUTS: INTEGER, FALSE IF THERE IS NO OCCURRENCE
        '''
        # first occurrence of uppercase
        if bl_uppercase == True:
            try:
                return list_.index(next(obj for obj in list_ if obj.isupper() == True))
            except StopIteration:
                return int_error
        # first occurrence of a given object
        elif obj_occurrence != None:
            for el in list_:
                if bl_audit == True:
                    print('BOOLEAN - MATCH STRING: {} {} {}'.format(
                        StrHandler().remove_diacritics(el),
                        StrHandler().remove_diacritics(obj_occurrence),
                        StrHandler().match_string_like(
                            StrHandler().remove_diacritics(el),
                            StrHandler().remove_diacritics(obj_occurrence))
                    ))
                if StrHandler().match_string_like(
                        StrHandler().remove_diacritics(el),
                        StrHandler().remove_diacritics(obj_occurrence)):
                    if bl_audit == True:
                        print('LIST INDEX: {}'.format(
                            list_.index(el)
                        ))
                    return list_.index(el)
            else:
                return int_error_obj_occurrence
        # last occurrence of an uppercase before a capitalized one
        elif bl_last_uppercase_before_capitalized == True:
            #   checking wheter the first occurrence is a capitalize or lower letter, in this
            #       case return the integer error - in this case return the string with no
            #       trims
            if bl_audit == True:
                print('RECEIVED LIST: {}'.format(list_))
                print('FIRST OBSERVATION TO ASSESS: {}'.format(list_[0]))
                print('BOOLEAN 1 - CAPITZALIDED: {}'.format(
                    StrHandler().is_capitalized(StrHandler().remove_diacritics(
                        list_[0])) == True
                ))
                print('BOOLEAN 2 - LOWER: {}'.format((StrHandler().remove_diacritics(
                    list_[0]).islower() == True)))
            if (StrHandler().is_capitalized(StrHandler().remove_diacritics(
                list_[0])) == True) or (StrHandler().remove_diacritics(
                    list_[0]).islower() == True):
                return int_error
            for i in range(len(list_) - 2):
                if (list_[i].replace(',', '').isupper() == True) and ((
                        StrHandler().is_capitalized(list_[i + 1].replace(',', '')) == True) or
                        (list_[i + 1].islower() == True)):
                    if bl_ignore_sole_letter == True:
                        if len(StrHandler().remove_non_alphanumeric_chars(list_[i])) == 1:
                            return i - 1
                        else:
                            return i
                    else:
                        return i
            else:
                return False
        # find first error to regex alphanumeric within a list
        elif bl_regex_alphanumeric_false == True:
            for i in range(len(list_)):
                if StrHandler().regex_match_alphanumeric(StrHandler().remove_diacritics(
                    list_[i].replace(
                        str_original_replace_1, str_result_replace).replace(
                            str_original_replace_2, str_result_replace)).strip()) == None:
                    return i
            else:
                return int_error
        else:
            raise Exception(
                'Neither boolean uppercase, nor object occurrence, were searched '
                + 'within the list for the first manifestation, please revisit the inputs')

    def get_list_until_invalid_occurrences(self, list_, list_invalid_values):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS
        '''
        # remove diacritcs from both lists
        list_ = [StrHandler().remove_diacritics(el) for el in list_]
        list_invalid_values = [StrHandler().remove_diacritics(el) for el in
                               list_invalid_values]
        # setting initial variables
        list_export = list()
        # looping through each element to find the first occurrence, and then break the loop
        for el in list_:
            if any([StrHandler().match_string_like(el, str_) == True for str_
                    in list_invalid_values]):
                break
            else:
                list_export.append(el)
        return list_export

    def first_numeric(self, list_):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS
        '''
        try:
            return next(iter([el for el in list_ if str(el).isnumeric()]))
        except StopIteration:
            return False

    def get_lower_upper_bound(self, sorted_list, value_to_put_in_between):
        '''
        REFERENCES: https://stackoverflow.com/questions/55895500/need-to-check-if-value-is-between-two-numbers-in-a-list
        DOCSTRING: LOWER AND UPPER BOUND OF DATA THAT A VALUE IS IN BETWEEN; IT CONSIDERS A
            LIST IN ASCENDING ORDER
        INPUTS: SORTED LIST AND VALUE TO BE IN BETWEEN OF DATA WITHIN THE LIST
        OUTPUTS JSON WITH LOWER AND UPPER BOUND
        '''
        # list index for for lower and upper bound
        if (value_to_put_in_between in sorted_list) and sorted_list[-1] != value_to_put_in_between:
            list_idx_lower_upper_bound = [bisect.bisect_left(sorted_list, value_to_put_in_between),
                                          bisect.bisect_left(sorted_list, value_to_put_in_between) + 1]
        else:
            list_idx_lower_upper_bound = [bisect.bisect_left(sorted_list, value_to_put_in_between) - 1,
                                          bisect.bisect_left(sorted_list, value_to_put_in_between)]
        # dictionary with responses
        if all(0 <= i <= len(sorted_list) for i in list_idx_lower_upper_bound):
            dict_message = {
                'lower_bound': sorted_list[list_idx_lower_upper_bound[0]],
                'upper_bound': sorted_list[list_idx_lower_upper_bound[1]]
            }
            return JsonFiles().send_json(dict_message)
        else:
            raise Exception('{} value is outside the bounds of {}'.format(
                value_to_put_in_between, sorted_list))

    def get_lower_mid_upper_bound(self, sorted_list, value_to_put_in_between):
        '''
        DOCSTRING: LOWER, MIDDLE AND UPPER BOUND OF DATA THAT A VALUE IS IN BETWEEN; IT CONSIDERS A
            LIST IN ASCENDING ORDER
        INPUTS: SORTED LIST AND VALUE TO BE IN BETWEEN OF DATA WITHIN THE LIST
        OUTPUTS JSON WITH LOWER, MIDDLE, UPPER BOUND AND BOOLEAN WITH END OF LIST
        '''
        # list index for for lower and upper bound
        if (value_to_put_in_between in sorted_list) and sorted_list[-1] != value_to_put_in_between:
            list_idx_lower_upper_bound = [bisect.bisect_left(sorted_list, value_to_put_in_between),
                                          bisect.bisect_left(sorted_list, 
                                                             value_to_put_in_between) + 1]
        else:
            list_idx_lower_upper_bound = [bisect.bisect_left(sorted_list, 
                                                             value_to_put_in_between) - 1,
                                          bisect.bisect_left(sorted_list, value_to_put_in_between)]
        # dictionary with responses
        if all(0 <= i <= len(sorted_list) for i in list_idx_lower_upper_bound) and \
                len(sorted_list) > 2:
            try:
                dict_message = {
                    'lower_bound': sorted_list[list_idx_lower_upper_bound[0]],
                    'middle_bound': sorted_list[list_idx_lower_upper_bound[1]],
                    'upper_bound': sorted_list[list_idx_lower_upper_bound[1] + 1],
                    'end_of_list': False
                }
            except:
                dict_message = {
                    'lower_bound': sorted_list[list_idx_lower_upper_bound[0] - 1],
                    'middle_bound': sorted_list[list_idx_lower_upper_bound[0]],
                    'upper_bound': sorted_list[list_idx_lower_upper_bound[1]],
                    'end_of_list': True
                }
            return JsonFiles().send_json(dict_message)
        else:
            raise Exception('{} value is outside the bounds of {}'.format(
                value_to_put_in_between, sorted_list))

    def closest_bound(self, sorted_list, value_to_put_in_between):
        '''
        DOCSTRING: CLOSEST BOUND TO A VALUE IN A LIST
        INPUTS: SORTED LIST, VALUE TO PUT IN BETWEEN
        OUTPUTS: VALUE
        '''
        return sorted_list[min(range(len(sorted_list)), key=lambda i:
                               abs(sorted_list[i] - value_to_put_in_between))]
    
    def closest_number_within_list(list_, number_):
        '''
        DOCSTRING: CLOSEST NUMBER TO NUMBER_ WITHIN A LIST
        INPUTS: LIST OF NUMBERS (NOT NECESSARILY SORTED) AND NUMBER K
        OUTPUTS: FLOAT/INTEGER
        '''
        return list_[min(range(len(list_)), key=lambda i: abs(list_[i]-number_))]

    def first_occurrence_like(self, list_, str_like):
        '''
        DOCSTRING: FIRST OCCURRENCE OF A MATCHING STRING WITHIN A LIST
        INPUTS: LIST AND STRING LIKE
        OUTPUTS: INTEGER
        '''
        return list_.index(next(x for x in list_ if StrHandler().match_string_like(
            x, str_like) == True))

    def remove_duplicates(self, list_interest):
        '''
        DOCSTRING: REMOVING DUPLICATES FROM A GIVEN LIST
        INPUTS: LIST
        OUTPUTS: LIST WITHOUT DUPLICATES
        '''
        return list(OrderedDict.fromkeys(list_interest))

    def nth_smallest_numbers(self, list_numbers, nth_smallest):
        '''
        DOCSTRING: RETURN THE NTH-SMALLEST NUMBERS FROM A LIST
        INPUTS: LIST NUMBERS
        OUTPUTS: NUMPY ARRAY
        '''
        # turning into a array
        array_numbers = np.array(list_numbers)
        # sort array
        array_numbers = np.sort(array_numbers)
        # returning the nth-smallest numnbers
        return array_numbers[0:nth_smallest]

    def extend_lists(self, *lists, bl_remove_duplicates=True):
        '''
        DOCSTRING: EXTEND N-LISTS AND REMOVE ITS DUPLICATES
        INPUTS: *ARGS WITH N-LISTS
        OUTPUTS: LIST
        '''
        # returning list with n-lists to append and remove duplicates
        list_extended_lists = list()
        # iterating through each list and appending to the final one
        for list_ in lists:
            list_extended_lists = chain(list_extended_lists, list_)
        # removing duplicates
        if bl_remove_duplicates == True:
            list_extended_lists = HandlingLists().remove_duplicates(list_extended_lists)
        else:
            list_extended_lists = list(list_extended_lists)
        # returning final list
        return list_extended_lists

    def chunk_list(self, list_to_chunk, str_character_divides_clients=' ',
                   int_chunk=150, bl_remove_duplicates=True):
        '''
        DOCSTRING: LIST TO CHUNK IN THE LIMIT SIZE
        INPUTS: LIST TO CHUNK, STRING CHARACTER TO DIVIDE CLIENT (DEFAULT), AND CHUNK (DEFAULT)
        OUTPUTS: LIST
        '''
        # setting variables
        list_chunked = list()
        # remove duplicates if is user's will
        if bl_remove_duplicates == True:
            list_to_chunk = HandlingLists().remove_duplicates(list_to_chunk)
        # creating chunks positions
        list_position_chunks = NumHandler().multiples(int_chunk, len(list_to_chunk))
        inf_limit = list_position_chunks[0]
        sup_limit = list_position_chunks[1]
        # checking wheter str_character_divides_clients is None, in this case append lists
        if str_character_divides_clients == None:
            return [list_to_chunk[x: x + int_chunk] for x in range(0, len(list_to_chunk), int_chunk)]
        # iterating through list to chunk, dividing in sublists with maximum size
        if len(list_position_chunks) > 2:
            for lim in list_position_chunks[2:]:
                #   narrowing query list
                if str_character_divides_clients == None:
                    list_chunked.append(list_to_chunk[inf_limit: sup_limit])
                else:
                    list_chunked.append(str_character_divides_clients.join(
                        list_to_chunk[inf_limit: sup_limit]))
                #   cutting limits
                inf_limit = sup_limit
                sup_limit = lim
            #   last append of sublist
            if str_character_divides_clients == None:
                list_chunked.append(list_to_chunk[
                    list_position_chunks[-2]: list_position_chunks[-1]])
            else:
                list_chunked.append(str_character_divides_clients.join(list_to_chunk[
                    list_position_chunks[-2]: list_position_chunks[-1]]))
        else:
            #   append list, if the size is inferior to chunk
            list_chunked.append(str_character_divides_clients.join(list_to_chunk[
                inf_limit: sup_limit]))
        #   removing duplicates
        list_chunked = HandlingLists().remove_duplicates(list_chunked)
        #   returning final result
        return list_chunked

    def cartesian_product(self, list_lists, int_break_n_n=None):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # setting variables
        list_export = list()
        # list of cartesian product of lists
        list_cartesian_product = list(product(*list_lists))
        # iterating through cartesian products of lists, if break in max values for tuples is different
        #   from none
        if int_break_n_n != None:
            for tup in list_cartesian_product:
                if (tup[:int_break_n_n] not in list_export) and (all([tup[:int_break_n_n][i] != tup[
                        :int_break_n_n][i - 1] for i in range(1, len(tup[:int_break_n_n]))])):
                    list_export.append(tup[:int_break_n_n])
            return list_export
        else:
            return list_cartesian_product

    def sort_alphanumeric(self, list_):
        '''
        REFERENCES: https://stackoverflow.com/questions/2669059/how-to-sort-alpha-numeric-set-in-python
        DOCSTRING: SORT ALPHANUMERIC DATA FROM LIST
        INPUTS:
        OUTPUTS:
        '''
        def convert(text): return int(text) if text.isdigit() else text
        def alphanum_key(key): return [convert(c)
                                       for c in re.split('([0-9]+)', key)]
        return sorted(list_, key=alphanum_key)

    def pairwise(self, iterable):
        '''
        REFERENCES: https://docs.python.org/3/library/itertools.html#itertools.pairwise
        DOCSTRING: RETURN SUCCESSIVE OVERLAPPING PAIRS TAKEN FROM THE INPUT ITERABLE
        INPUTS: ITERABLE (LIST)
        OUTPUTS: LIST OF TUPLES
        '''
        a, b = tee(iterable)
        next(b, None)
        return list(zip(a, b))

    def discard_from_list(self, list_, list_items_remove):
        '''
        DOCSTRING: DISCARD-LIKE, FROM SETS, TO LISTS
        INPUTS: LIST AND LIST OF ELEMENTS TO REMOVE
        OUTPUTS: -
        '''
        for item in list_items_remove:
            if item in list_:
                list_.remove(item)
        return list_

    def absolute_frequency(self, list_):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return Counter(list_)

    def flatten_list(self, list_):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return [x for xs in list_ for x in xs]

class PriorityQueue:
    '''
    REFERENCES: PYTHON COOKBOOK - DAVID BEASZLEY, BRIAN K. JONES
    DOCSTRING: CREATE A PRIORITY QUEUE LIST
    INPUTS: -
    OUTPUTS: -
    '''

    def __init__(self):
        self._queue = []
        self._index = 0

    def push(self, item, priority):
        '''
        DOCSTRING: PUSH ITEMS TO LIST
        INPUTS: ITEM, PRIORITY
        OUTPUTS: -
        '''
        heappush(self._queue, (-priority, self._index, item))
        self._index += 1

    def pop(self):
        '''
        DOCSTRING: REMOVE LAST ITEM FROM LIST
        INPUTS: -
        OUTPUTS: OBJECT
        '''
        return heappop(self._queue)[-1]

# sorted_list = [10, 30, 52, 73, 90, 113, 133, 154, 175, 197, 219, 240, 303, 365, 427, 492, 554, 617, 678,
#                742, 803, 864, 927, 993, 1057, 1118, 1179, 1245, 1310, 1432, 1560, 1811, 2060, 2310, 2563, 3066, 3566]
# value_needle = 2564
# print(HandlingLists().get_lower_mid_upper_bound(sorted_list, value_needle))
# print(sorted_list)
# print(value_needle)
# # output
# {'lower_bound': 2563, 'middle_bound': 3066, 'upper_bound': 3566, 'end_of_list': True}

# sorted_list = [10, 30, 52, 73, 90, 113, 133, 154, 175, 197, 219, 240, 303, 365, 427, 492, 554, 617, 678,
#                742, 803, 864, 927, 993, 1057, 1118, 1179, 1245, 1310, 1432, 1560, 1811, 2060, 2310, 2563, 3066, 3566]
# value_needle = 3100
# print(HandlingLists().get_lower_upper_bound(sorted_list, value_needle))
# print(sorted_list)
# print(value_needle)
# # output
# {'lower_bound': 3066, 'upper_bound': 3566}
# list_ = ['Bacterias', 'mesofilas', 'aerobias', 'e', 'anaerobias',
#          'facultativas', '–', 'ISO', '4833-2.', '2013', 'FDA.', 'BAM.', 'Ch']
# print(HandlingLists().get_first_occurrence_within_list(
#     list_, bl_last_uppercase_before_capitalized=True))
# list_ = ['SWAB', 'DE', 'CARCAÇAS', 'S.', 'aureus', '–',
#          'Determinação', 'quantitativa', 'pela', 'técnica', 'de']
# print(HandlingLists().get_first_occurrence_within_list(
#     list_, bl_last_uppercase_before_capitalized=True))
